import numpy as np
import argparse
import os
import sys
from schwimmbad import JoblibPool
from bilby.core.utils import check_directory_exists_and_if_not_mkdir

sys.path.append(
    os.path.join(
        os.path.abspath(os.path.dirname(__file__)), "../../calibration"
    )
)
from models import *
from metrics import *
from parameters import parameters_random

from SEOBNRv5_wrapper import SEOBNRv5_wrapper


def attachment_time(q, chi1, chi2):
    
    omega0 = 0.015

    model = SEOBNRv5_wrapper(
        q,
        chi1,
        chi2,
        omega0,
    )
    
    t_attach = model.t_attach

    return (
        q,
        chi1,
        chi2,
        t_attach
    )


def process_one_case(input):
    q, chi1, chi2 = input
    q, chi1, chi2, t_attach = attachment_time(q, chi1, chi2)
    return np.array([q, chi1, chi2, t_attach])


if __name__ == "__main__":

    p = argparse.ArgumentParser(
        description="Check attachment time smoothness across parameter space"
    )
    p.add_argument(
        "--points", type=int, help="Number of points", default="5000"
    )
    p.add_argument(
        "--q-max", type=float, help="Upper limit on q", default="20.0"
    )
    p.add_argument(
        "--chi-max", type=float, help="Upper limit on |chi|", default="0.95"
    )
    p.add_argument(
        "--name", type=str, help="Name of the output file", default="attachment_time.dat"
    )
    p.add_argument(
        "--plots", action="store_true", help="Make diagnostic plots", default=True
    )
    args = p.parse_args()

    qarr, chi1arr, chi2arr = parameters_random(args.points, 1.0, args.q_max, -args.chi_max, args.chi_max, -args.chi_max, args.chi_max)

    lst = [(a, b, c) for a, b, c in zip(qarr, chi1arr, chi2arr)]

    with JoblibPool(24) as pool:
        all_means = pool.map(process_one_case, lst)

    all_means = np.array(all_means)

    np.savetxt(args.name, all_means)

    ### Plots ###

    if args.plots:
        
        import matplotlib
        import matplotlib.pyplot as plt

        plt_dir = './plots'
        check_directory_exists_and_if_not_mkdir(plt_dir)

        res_path = args.name
        res = np.loadtxt(res_path)

        q = res[:,0]
        chi1 = res[:,1]
        chi2 = res[:,2]
        t_attach = res[:,3]

        nu = q/(1+q)**2
        m1 = q/(1+q)
        m2 = 1/(1 + q)
        ap = m1*chi1+m2*chi2
        am = m1*chi1-m2*chi2

        # Attachment time across parameter space - (q, chieff)
        plt.scatter(q, ap, c=t_attach,linewidths=1,norm=matplotlib.colors.LogNorm())
        plt.ylabel('$\chi_{\mathrm{eff}}$')
        plt.xlabel('$q$')
        cbar=plt.colorbar()
        cbar.set_label('$t_{\mathrm{attach}}/M (\Omega_0 = 0.015)$')
        plt.savefig(f'{plt_dir}/t_attachment.png', bbox_inches = 'tight', dpi = 300)
        plt.close()



# cython: language_level=3, profile=True, linetrace=True, binding=True

"""
Contains the functions needed for computing the "analytic" post-adiabatic dynamics
"""

cimport cython
from typing import Dict
import numpy as np
#import findiff
from scipy.integrate import solve_ivp, ode
from .initial_conditions_aligned_opt import computeIC_opt as computeIC


from pygsl_lite import  roots, errno
from scipy import optimize
from scipy.interpolate import CubicSpline, InterpolatedUnivariateSpline
from scipy import integrate

from lalinference.imrtgr import nrutils

from .integrate_ode import compute_dynamics_opt as compute_dynamics
from .integrate_ode import augment_dynamics
from pyseobnr.eob.utils.containers cimport EOBParams
from pyseobnr.eob.hamiltonian.Hamiltonian_C cimport Hamiltonian_C
from pyseobnr.eob.waveform.waveform cimport RadiationReactionForce
from libc.math cimport log, sqrt, exp, abs,fabs, tgamma,sin,cos, tanh, sinh, asinh


from pyseobnr.eob.dynamics.postadiabatic_C cimport fin_diff_derivative, cumulative_integral, Kerr_ISCO, compute_adiabatic_solution


@cython.wraparound(False)
@cython.boundscheck(False)
@cython.cdivision(True)
@cython.nonecheck(False)
@cython.initializedcheck(False)
@cython.profile(True)
@cython.linetrace(True)
cpdef double pr_eqn(
    double pr_sol,
    double r,
    double pphi,
    double dpphi_dr,
    Hamiltonian_C H,
    RadiationReactionForce RR,
    double chi_1,
    double chi_2,
    double m_1,
    double m_2,
    EOBParams params,
    double[::1] q,
    double[::1] p
):
    """
    Compute the value of pr at odd PA orders.
    This is done by evaluating Eq.(C7) of SEOBNRv5 theory doc
    at a given radial grid point. See  DCC:T2300060
    """
    q[0] = r
    p[0] = pr_sol
    p[1] = pphi
    cdef double dAdr,dBnpdr,dBnpadr,dxidr,dQdr,dQdprst,dHodddr,xi,omega,omega_circ
    cdef double A,Bnp,Bnpa,Q,Heven,Hodd,H_val
    cdef double aux_derivs[7]
    cdef double nu = params.p_params.nu
    aux_derivs[:] = H.auxderivs(q, p, chi_1, chi_2, m_1, m_2)

    dAdr = aux_derivs[0]
    dBnpdr = aux_derivs[1]
    dBnpadr = aux_derivs[2]
    dxidr = aux_derivs[3]
    dQdr = aux_derivs[4]
    dQdprst = aux_derivs[5]
    dHodddr = aux_derivs[6]

    H_val,xi,A,Bnp,Bnpa,Q,Heven,Hodd = H(q,p,chi_1,chi_2,m_1,m_2,verbose=True)
    omega = H.omega(q,p,chi_1,chi_2,m_1,m_2)

    params.dynamics.p_circ[1] = pphi
    omega_circ = H.omega(q,params.dynamics.p_circ,chi_1,chi_2,m_1,m_2)

    cdef (double,double) flux = RR.RR(q, p, omega,omega_circ,H_val,params)


    cdef double result =  xi/(2*(1+Bnp))*(flux[1]/dpphi_dr*2*nu*H_val*Heven/A - xi*dQdprst)


    return result


@cython.profile(True)
@cython.linetrace(True)
cpdef compute_pr(
    double[:] r,
    double[:] pr,
    double[:] pphi,
    Hamiltonian_C H,
    RadiationReactionForce RR,
    double chi_1,
    double chi_2,
    double m_1,
    double m_2,
    double[::1] q,
    double[::1] p,
    EOBParams params=None,
):
    """
    Compute the value to pr at odd PA orders.
    This is done by evaluating  Eq(C7) of
    SEOBNRv5 theory doc at every radial grid point.
    See DCC:T2300060
    """
    cdef double[:] dpphi_dr = - fin_diff_derivative(r, pphi)
    cdef int i
    for i in range(r.shape[0]):

        pr[i]= pr_eqn(pr[i],r[i], pphi[i], dpphi_dr[i],
                H, RR, chi_1, chi_2, m_1, m_2,
                params,q,p)

    return pr

@cython.wraparound(False)
@cython.boundscheck(False)
@cython.cdivision(True)
@cython.nonecheck(False)
@cython.initializedcheck(False)
@cython.profile(True)
@cython.linetrace(True)
cpdef double pphi_eqn(
    double pphi_sol,
    double r,
    double pr,
    double dpr_dr,
    Hamiltonian_C H,
    RadiationReactionForce RR,
    double chi_1,
    double chi_2,
    double m_1,
    double m_2,
    EOBParams params,
    double[::1] q,
    double[::1] p
):
    """
    Compute value of pphi at a given radial grid point.
    This is done by solving Eq.(C8) from SEOBNRv5
    theory doc. See DCC:T2300060.
    """
    q[0] = r
    p[0] = pr
    p[1] =  pphi_sol
    cdef double nu = params.p_params.nu
    params.dynamics.p_circ[1] =  pphi_sol

    cdef double dAdr,dBnpdr,dBnpadr,dxidr,dQdr,dQdprst,dHOdddr,xi,omega,omega_circ,result
    cdef double A,Bnp,Bnpa,Q,Heven,Hodd,H_val
    cdef double aux_derivs[7]

    aux_derivs[:] = H.auxderivs(q, p, chi_1, chi_2, m_1, m_2)

    dAdr = aux_derivs[0]
    dBnpdr = aux_derivs[1]
    dBnpadr = aux_derivs[2]
    dxidr = aux_derivs[3]
    dQdr = aux_derivs[4]
    dQdprst = aux_derivs[5]
    dHodddr = aux_derivs[6]

    H_val,xi,A,Bnp,Bnpa,Q,Heven,Hodd = H(q,p,chi_1,chi_2,m_1,m_2,verbose=True)
    omega = H.omega(q,p,chi_1,chi_2,m_1,m_2)

    params.dynamics.p_circ[1] = pphi_sol
    omega_circ = H.omega(q,params.dynamics.p_circ,chi_1,chi_2,m_1,m_2)

    cdef double drdt = A/(2*nu*H_val*Heven)*(2*pr/xi*(1+Bnp)+xi*dQdprst)

    cdef (double,double) flux = RR.RR(q, p, omega,omega_circ,H_val,params)

    cdef double pr2 = pr*pr
    cdef double xi2 = xi*xi
    cdef double xi3 = xi2*xi
    cdef double r2 = r*r
    cdef double r3 = r2*r
    cdef double ap = m_1*chi_1+m_2*chi_2
    cdef double ap2 = ap*ap

    cdef double C_2 = dAdr*(1/r2+ap2/r2*Bnpa)+A*(-2/r3-2*ap**2/r3*Bnpa+ap2/r2*dBnpadr)
    cdef double C_1 = dHodddr/pphi_sol*2*Heven
    cdef double C_0_pt1 = dAdr*(1+pr2/xi2*Bnp+pr2/xi2+Q) + A*(-2*pr2/xi3*Bnp*dxidr+pr2/xi2*dBnpdr-2*pr2/xi3*dxidr+dQdr)
    cdef double tmp = 2*Heven*nu*H_val/xi
    cdef double C_0_pt2 = dpr_dr*drdt*tmp - pr/pphi_sol*tmp*flux[1]
    cdef double C_0 = C_0_pt1+C_0_pt2

    D = C_1*C_1 - 4*C_2*C_0

    if D<0 and fabs(D)<1e-10:
        D = 0.0
    if D<0:
        raise ValueError
    result = (-C_1-sqrt(D))/(2*C_2)
    return result



@cython.profile(True)
@cython.linetrace(True)
cpdef compute_pphi(
    double[:] r,
    double[:] pr,
    double[:] pphi,
    Hamiltonian_C H,
    RadiationReactionForce RR,
    double chi_1,
    double chi_2,
    double m_1,
    double m_2,
    double[::1] q,
    double[::1] p,
    EOBParams params=None
):
    """
    Compute value of pphi at even PA orders.
    This is done by solving Eq.(C8) from SEOBNRv5
    theory doc at every radial grid point.
    See DCC:T2300060
    """

    cdef double[:] dpr_dr = - fin_diff_derivative(r, pr)
    cdef int i
    for i in range(r.shape[0]):

        pphi[i]  = pphi_eqn(pphi[i],r[i], pr[i], dpr_dr[i], H, RR, chi_1, chi_2, m_1, m_2, params,q,p)

    return pphi

@cython.profile(True)
@cython.linetrace(True)
cpdef compute_postadiabatic_solution(
    double[:] r,
    double[:] pphi,
    Hamiltonian_C H,
    RadiationReactionForce RR,
    double chi_1,
    double chi_2,
    double m_1,
    double m_2,
    double[::1] q,
    double[::1] p,
    double tol=1e-12,
    int order=8,
    EOBParams params=None,
):
    """
    Compute the post-adiabatic values
    of pr and pphi iteratively up to the given PA order.
    """
    pr = np.zeros(r.size)
    cdef int n
    cdef double tol_current
    for n in range(1, order+1):
        parity = n % 2
        if parity:
            pr = compute_pr(
                r,
                pr,
                pphi,
                H,
                RR,
                chi_1,
                chi_2,
                m_1,
                m_2,
                q,
                p,
                params=params,
            )
        else:
            pphi = compute_pphi(
                r,
                pr,
                pphi,
                H,
                RR,
                chi_1,
                chi_2,
                m_1,
                m_2,
                q,
                p,
                params=params,
            )

    return pr, pphi

@cython.profile(True)
@cython.linetrace(True)
cpdef compute_postadiabatic_dynamics(
    double omega0,
    Hamiltonian_C H,
    RadiationReactionForce RR,
    double chi_1,
    double chi_2,
    double m_1,
    double m_2,
    double tol=1e-12,
    EOBParams params=None,
    int order=8
):
    """Compute the dynamics using PA procedure starting from omega0

    Args:
        omega0 (float): The starting *orbital* frequency in geomtric units
        H (Hamiltonian): The Hamiltonian object to use
        RR (function): The RR function to use
        chi_1 (float): z-component of the primary spin
        chi_2 (float): z-component of the secondary spin
        m_1 (float): mass of the primary
        m_2 (float): mass of the secondary

    Returns:
        np.array: The dynamics results, as (t,q,p)
    """
    cdef double r0
    r0, _, _ = computeIC(
        omega0,
        H,
        RR,
        chi_1,
        chi_2,
        m_1,
        m_2,
        params=params,
    )

    cdef double chi_eff = m_1*chi_1+m_2*chi_2
    cdef double nu = m_1*m_2/(m_1+m_2)**2
    cdef double r_final_prefactor = 2.7+chi_eff*(1-4.*nu)
    r_ISCO, _ = Kerr_ISCO(chi_1, chi_2, m_1, m_2)
    cdef double r_final = max(10,r_final_prefactor * r_ISCO)
    cdef double r_switch_prefactor = 1.6
    cdef double r_switch = r_switch_prefactor * r_ISCO

    cdef double dr0 = 0.2
    cdef int r_size = int(np.ceil((r0 - r_final) / dr0))

    if r_size <= 4 or r0<=11.5:
        raise ValueError
    elif r_size < 10:
        r_size = 10

    r, _ = np.linspace(r0, r_final, num=r_size, endpoint=True, retstep=True)
    cdef double[::1] q = np.zeros(2)
    cdef double[::1] p = np.zeros(2)
    pphi = compute_adiabatic_solution(r, H, chi_1, chi_2, m_1, m_2,q,p, tol=tol,)

    pr, pphi = compute_postadiabatic_solution(
        r,
        pphi,
        H,
        RR,
        chi_1,
        chi_2,
        m_1,
        m_2,
        q,
        p,
        tol=tol,
        order=order,
        params=params,
    )

    dt_dr = np.zeros(r.shape[0])
    dphi_dr = np.zeros(r.shape[0])
    cdef int i
    cdef double dH_dpr,dH_dpphi,csi
    cdef double dyn[6]
    for i in range(r.shape[0]):
        q = np.array([r[i], 0])
        p = np.array([pr[i], pphi[i]])

        dyn[:] = H.dynamics(q, p, chi_1, chi_2, m_1, m_2)
        dH_dpr = dyn[2]
        dH_dpphi = dyn[3]


        csi = dyn[5]
        dt_dr[i] = 1 / (dH_dpr * csi)
        dphi_dr[i] = dH_dpphi / (dH_dpr * csi)


    t = cumulative_integral(r, dt_dr)

    phi = cumulative_integral(r, dphi_dr)

    postadiabatic_dynamics = np.c_[t, r, phi, pr, pphi]




    return postadiabatic_dynamics


@cython.boundscheck(False)
@cython.nonecheck(False)
@cython.initializedcheck(False)
@cython.profile(True)
@cython.linetrace(True)
cpdef compute_combined_dynamics(
    double omega0,
    Hamiltonian_C H,
    RadiationReactionForce RR,
    double chi_1,
    double chi_2,
    double m_1,
    double m_2,
    double tol=1e-12,
    EOBParams params=None,
    double step_back=50,
    str backend="ode",
    int PA_order=8,
    double r_stop = -1
):
    """
    Compute the full inspiral dynamics by combining PA + ODE
    integration. If the PA procedure fails (e.g. because the
    inspiral is too short), the code falls back to ODE
    integration.
    """
    try:
        postadiabatic_dynamics = compute_postadiabatic_dynamics(
            omega0,
            H,
            RR,
            chi_1,
            chi_2,
            m_1,
            m_2,
            tol=tol,
            params=params,
            order=PA_order
        )
        PA_success = True
        ode_y_init = postadiabatic_dynamics[-1, 1:]
    except Exception as e:

        PA_success = False
        ode_y_init = None


    #print("PA_success: ", PA_success)
    #print("ode_y_init: ", ode_y_init)
    ode_dynamics_low, ode_dynamics_high = compute_dynamics(
        omega0,
        H,
        RR,
        chi_1,
        chi_2,
        m_1,
        m_2,
        rtol=1e-11,
        atol=1e-12,
        backend=backend,
        params=params,
        step_back=step_back,
        max_step=0.5,
        min_step=1.0e-9,
        y_init=ode_y_init,
        r_stop=r_stop
    )

    if PA_success is True:
        postadiabatic_dynamics = augment_dynamics(postadiabatic_dynamics,chi_1,chi_2,m_1,m_2,H)
        ode_dynamics_low[:, 0] += postadiabatic_dynamics[-1, 0]
        ode_dynamics_high[:, 0] += postadiabatic_dynamics[-1, 0]
        combined_dynamics = np.vstack(
            (postadiabatic_dynamics[:-1], ode_dynamics_low)
        )
    else:
        combined_dynamics = ode_dynamics_low


    return combined_dynamics, ode_dynamics_high



#!/usr/bin/env python3
"""
Abstract Hamiltonian classes
"""
from abc import abstractmethod


class Hamiltonian:
    """Aligned-spin Hamiltonian class"""

    def __init__(self):
        pass

    @abstractmethod
    def __call__(self, q, p, chi_1, chi_2, **kwargs):
        """Evaluate the Hamiltonian"""
        pass

    @abstractmethod
    def grad(self):
        """Return the gradient of the Hamiltonian with respect to dynamical vars"""
        pass

    @abstractmethod
    def hessian(self):
        """Return the Hessian of the Hamiltonian with respect to dynamical vars. Needed for IC"""
        pass


class Hamiltonian_v5PHM:
    """Precessing-spin Hamiltonian class"""

    def __init__(self):
        pass

    @abstractmethod
    def __call__(
        self, q, p, chi1_v, chi2_v, m_1, m_2, chi_1, chi_2, chiL_1, chiL_2, **kwargs
    ):
        """Evaluate the Hamiltonian"""
        pass

    @abstractmethod
    def grad(self):
        """Return the gradient of the Hamiltonian with respect to dynamical vars"""
        pass

    @abstractmethod
    def hessian(self):
        """Return the Hessian of the Hamiltonian with respect to dynamical vars. Needed for IC"""
        pass



#!/usr/bin/env python3
import logging
import os
from abc import ABC
from typing import Any, Dict

import h5py
import lal
import lalsimulation as lalsim
import numpy as np
import scri
import sxs
from rich.logging import RichHandler
from rich.traceback import install
from scipy.interpolate import CubicSpline, InterpolatedUnivariateSpline

from .default_settings import *
from pyseobnr.models.model import Model

logger = logging.getLogger(__name__)
logger.addHandler(RichHandler(rich_tracebacks=True, markup=True))
logger.setLevel("INFO")
install()


class NRModel_LVC(Model):
    """Represents a numerical relativity waveform as evaluated through the LAL
    NR interface. Requires data in the LVCNR format.
    """

    def __init__(self, data_file: str, settings: Dict[Any, Any] = None) -> None:
        self.data_file = data_file
        self.settings = default_NR_LVC_settings()
        # If we were given settings, override the defaults
        if settings is not None:
            self.settings.update(**settings)

        # Deal with debug mode
        self.debug = self.settings.get("debug", False)
        if self.debug:
            logger.setLevel("DEBUG")

        with h5py.File(self.data_file, "r") as f:
            # We rescale the masses so that they add to 1!
            self.m_1 = f.attrs["mass1"] / (f.attrs["mass1"] + f.attrs["mass2"])
            self.m_2 = f.attrs["mass2"] / (f.attrs["mass1"] + f.attrs["mass2"])
            # Starting frequency in Hz at 1 solar mass total mass
            Mf_lower = f.attrs["f_lower_at_1MSUN"]

        self.q = self.m_1 / self.m_2
        self.M = self.settings["M"]

        f_lower = Mf_lower / (self.m_1 + self.m_2)
        (
            s1x,
            s1y,
            s1z,
            s2x,
            s2y,
            s2z,
        ) = lalsim.SimInspiralNRWaveformGetSpinsFromHDF5File(
            f_lower,
            1,
            self.data_file,
        )

        params = lal.CreateDict()
        lalsim.SimInspiralWaveformParamsInsertNumRelData(params, self.data_file)

        self.distance = self.settings["distance"]

        self.dt = self.settings["dt"]
        self.delta_T = self.dt / (self.M * lal.MTSUN_SI)

        self.f_start = Mf_lower / self.M
        self.f_ref = Mf_lower / self.M
        self.omega0 = self.f_start * self.M * lal.MTSUN_SI * np.pi

        ell_max = 8
        mode_array = lalsim.SimInspiralCreateModeArray()
        for ell in range(2, ell_max + 1):
            for m in range(-ell, ell + 1):
                mode_array = lalsim.SimInspiralModeArrayActivateMode(mode_array, ell, m)

        _, hlm = lalsim.SimInspiralNRWaveformGetHlms(
            self.dt,
            self.m_1 * self.M * lal.MSUN_SI,
            self.m_2 * self.M * lal.MSUN_SI,
            self.distance,
            self.f_start,
            self.f_ref,
            s1x,
            s1y,
            s1z,
            s2x,
            s2y,
            s2z,
            self.data_file,
            mode_array,
        )
        self.chi1 = np.array([s1x, s1y, s1z])
        self.chi2 = np.array([s2x, s2y, s2z])
        scaling_factor = self.M * lal.MRSUN_SI / self.distance

        self.waveform_modes = {}
        for mode in self.settings["modes"]:
            # Note the  rescaling of the amplitude from SI to geometric units
            self.waveform_modes[f"{mode[0]},{mode[1]}"] = (
                lalsim.SphHarmTimeSeriesGetMode(hlm, mode[0], mode[1]).data.data
                / scaling_factor
            )

        # We are guaranteed that the (2,2) mode is present
        tmp = lalsim.SphHarmTimeSeriesGetMode(hlm, 2, 2)
        # The shift
        shift = 0 * (tmp.epoch.gpsSeconds + tmp.epoch.gpsNanoSeconds / 1e9)
        # Rescale the time
        time_array = (np.arange(0, len(tmp.data.data)) * self.dt + shift) / (
            self.M * lal.MTSUN_SI
        )
        self.t = time_array

        """
        Compute and store the energy and momentum fluxes
        """
        energy_flux_sum = 0
        momentum_flux_sum = 0

        for ell in range(2, ell_max + 1):
            for m in range(-ell, ell + 1):
                h_lm = (
                    lalsim.SphHarmTimeSeriesGetMode(hlm, ell, m).data.data
                    / scaling_factor
                )
                h_lm_interp = CubicSpline(self.t, h_lm)
                dh_lm_dt = h_lm_interp.derivative()(self.t)

                energy_flux_sum += np.abs(dh_lm_dt) ** 2
                momentum_flux_sum += -m * np.imag(np.conjugate(dh_lm_dt) * h_lm)

        self.energy_flux = energy_flux_sum / (16 * np.pi)
        self.momentum_flux = momentum_flux_sum / (16 * np.pi)

    def __call__(self):
        """
        Since the modes are computed at initialization, don't need to do anything here
        """
        pass


class NRModel_SXS(Model):
    """Represents a numerical relativity waveform from SXS catalog.
    This requires the scri package.
    """

    def __init__(self, data_file: str, settings: Dict[Any, Any] = None) -> None:
        self.data_file = data_file
        self.settings = default_NR_SXS_settings()
        # If we were given settings, override the defaults
        if settings is not None:
            self.settings.update(**settings)

        # Deal with debug mode
        self.debug = self.settings.get("debug", False)
        if self.debug:
            logger.setLevel("DEBUG")

        self.delta_T = self.settings.get("dt", 0.1)
        waveform = scri.SpEC.read_from_h5(
            f"{data_file}/Extrapolated_N{self.settings['extrap_order']}.dir"
        )
        t_new = np.arange(waveform.t[0], waveform.t[-1], self.delta_T)
        waveform = waveform.interpolate(t_new)
        # Load the metadata. Note: assume it is in the same directory as the h5 file
        path = os.path.dirname(data_file)
        metadata = sxs.load(f"{path}/metadata.json")

        self.m1 = metadata.reference_mass1
        self.m2 = metadata.reference_mass2
        self.m_1 = self.m1
        self.m_2 = self.m2
        self.M = self.m1 + self.m2
        self.q = self.m1 / self.m2
        self.chi1 = metadata.reference_dimensionless_spin1
        self.chi2 = metadata.reference_dimensionless_spin2
        self.tr = metadata.reference_time

        self.omega0 = np.linalg.norm(metadata.initial_orbital_frequency)
        self.waveform_modes = {}
        ell_max = self.settings.get("ell_max", 8)
        idx_ok = np.where(waveform.t >= self.tr)
        for ell in range(2, ell_max + 1):
            for m in range(-ell, ell + 1):
                idx = waveform.index(ell, m)
                self.waveform_modes[f"{ell},{m}"] = waveform.data[idx_ok][:, idx]

        self.t = waveform.t[idx_ok]

        """
        Compute and store the energy and momentum fluxes
        """
        energy_flux_sum = 0
        momentum_flux_sum = 0

        for ell in range(2, ell_max + 1):
            for m in range(-ell, ell + 1):
                h_lm = self.waveform_modes[f"{ell},{m}"].copy()
                h_lm_interp = CubicSpline(self.t, h_lm)
                dh_lm_dt = h_lm_interp.derivative()(self.t)

                energy_flux_sum += np.abs(dh_lm_dt) ** 2
                momentum_flux_sum += -m * np.imag(np.conjugate(dh_lm_dt) * h_lm)

        self.energy_flux = energy_flux_sum / (16 * np.pi)
        self.momentum_flux = momentum_flux_sum / (16 * np.pi)

    def __call__(self):
        """
        Since the modes are computed at initialization, don't need to do anything here
        """
        pass


class NRModel_RIT(Model):
    """Represents a numerical relativity waveform from RIT catalog.
    This requires the sxs package.
    """

    def __init__(self, data_file: str, settings: Dict[Any, Any] = None) -> None:
        self.data_file = data_file
        self.settings = default_NR_RIT_settings()
        # If we were given settings, override the defaults
        if settings is not None:
            self.settings.update(**settings)

        # Deal with debug mode
        self.debug = self.settings.get("debug", False)
        if self.debug:
            logger.setLevel("DEBUG")

        # Load the metadata. Note: assume it is in the same directory as the h5 file
        path = os.path.dirname(data_file)
        metadata = sxs.load(f"{path}/metadata.json")

        # Note, for RIT m_2>=m_1. But we always have m_1>=m_2
        self.m1 = metadata.relaxed_mass2
        self.m2 = metadata.relaxed_mass1
        self.q = self.m1 / self.m2
        if "relaxed_chi1x" in metadata.keys():
            chi1 = np.array(
                [metadata.relaxed_chi2x, metadata.relaxed_chi2y, metadata.relaxed_chi2z]
            )
            chi2 = np.array(
                [metadata.relaxed_chi1x, metadata.relaxed_chi1y, metadata.relaxed_chi1z]
            )
        else:
            chi1 = np.array([0.0, 0.0, metadata.relaxed_chi2z])
            chi2 = np.array([0.0, 0.0, metadata.relaxed_chi1z])
        self.chi1 = chi1
        self.chi2 = chi2
        # FIXME: this is what was done for SEOBNRv4, but it probably makes more sense
        # to use the relaxed frequency
        self.omega0 = metadata.freq_start_22 * np.pi
        self.waveform_modes = {}
        ell_min, ell_max = 2, 4
        with h5py.File(self.data_file, "r") as f:
            t = f["NRTimes"][:]
            for ell in range(ell_min, ell_max + 1):
                for m in range(-ell, ell + 1):
                    amp = InterpolatedUnivariateSpline(
                        f[f"amp_l{ell}_m{m}/X"][:],
                        f[f"amp_l{ell}_m{m}/Y"][:],
                        k=f[f"amp_l{ell}_m{m}/deg"][()],
                    )(t)
                    phase = InterpolatedUnivariateSpline(
                        f[f"phase_l{ell}_m{m}/X"][:],
                        f[f"phase_l{ell}_m{m}/Y"][:],
                        k=f[f"phase_l{ell}_m{m}/deg"][()],
                    )(t)
                    self.waveform_modes[f"{ell},{m}"] = amp * np.exp(1j * phase)

        self.t = t - t[0]
        self.tr = metadata.relaxed_time

        """
        Compute and store the energy and momentum fluxes
        """
        energy_flux_sum = 0
        momentum_flux_sum = 0

        for ell in range(2, ell_max + 1):
            for m in range(-ell, ell + 1):
                h_lm = self.waveform_modes[f"{ell},{m}"].copy()
                h_lm_interp = CubicSpline(self.t, h_lm)
                dh_lm_dt = h_lm_interp.derivative()(self.t)

                energy_flux_sum += np.abs(dh_lm_dt) ** 2
                momentum_flux_sum += -m * np.imag(np.conjugate(dh_lm_dt) * h_lm)

        self.energy_flux = energy_flux_sum / (16 * np.pi)
        self.momentum_flux = momentum_flux_sum / (16 * np.pi)

    def __call__(self):
        """
        Since the modes are computed at initialization, don't need to do anything here
        """
        pass


class SEOBNRv4HM_LAL(Model):
    """
    Represents an SEOBNRv4HM waveform generated through LAL.
    It generates the following modes:
    (2,2), (3,3), (4,4), (5,5), (2,1)

    Args:
        q (float): Mass ratio m1/m2 >= 1
        chi_1 (float): z component of the dimensionless spin of primary
        chi_2 (float): z component of the dimensionless spin of secondary
        omega0 (float): Initial orbital frequency, in geomtric units
        settings (Dict[Any, Any], optional): The settings. Defaults to None.
    """

    def __init__(
        self,
        q: float,
        chi_1: float,
        chi_2: float,
        omega0: float,
        coeffs: Dict[str, Any] = None,
        settings: Dict[Any, Any] = None,
    ) -> None:
        self.settings = default_SEOBNRv4HM_settings()
        # If we were given settings, override the defaults
        if settings is not None:
            self.settings.update(**settings)

        self.M = self.settings["M"]
        self.q = q
        m_1 = q / (1 + q)
        m_2 = 1 - m_1

        self.chi_1 = chi_1
        self.chi_2 = chi_2

        self.dt = self.settings["dt"]
        self.delta_T = self.dt / (self.M * lal.MTSUN_SI)
        # Starting *orbital* frequency
        self.omega0 = omega0
        self.f0 = self.omega0 / (self.M * lal.MTSUN_SI * np.pi)
        self.distance = self.settings["distance"]

        # Component masses in solar masses
        self.m1_msun = m_1 * self.M
        self.m2_msun = m_2 * self.M

        # NQC coeffs
        self.nqcCoeffsInput = lal.CreateREAL8Vector(10)

        # SEOBNRv4HM is numbered 41
        self.wf_version = 41

        # Debug mode?
        self.debug = self.settings.get("debug", False)

    def __call__(self):
        self._evaluate_model()

    def _evaluate_model(self):
        """
        Evaluate the model
        """
        pars = {
            "phiC": 0.0,
            "deltaT": self.dt,
            "m1_SI": self.m1_msun * lal.MSUN_SI,
            "m2_SI": self.m2_msun * lal.MSUN_SI,
            "distance": self.distance,
            "inclination": 0.0,
            "lambda2Tidal1": 0.0,
            "lambda2Tidal2": 0.0,
            "omega02Tidal1": 0.0,
            "omega02Tidal2": 0.0,
            "lambda3Tidal1": 0.0,
            "lambda3Tidal2": 0.0,
            "omega03Tidal1": 0.0,
            "omega03Tidal2": 0.0,
            "quadparam1": 1.0,
            "quadparam2": 1.0,
        }

        sphtseries, dyn, dynHI = lalsim.SimIMRSpinAlignedEOBModes(
            pars["deltaT"],
            pars["m1_SI"],
            pars["m2_SI"],
            self.f0,
            pars["distance"],
            self.chi_1,
            self.chi_2,
            self.wf_version,
            pars["lambda2Tidal1"],
            pars["lambda2Tidal2"],
            pars["omega02Tidal1"],
            pars["omega02Tidal2"],
            pars["lambda3Tidal1"],
            pars["lambda3Tidal2"],
            pars["omega03Tidal1"],
            pars["omega03Tidal2"],
            pars["quadparam1"],
            pars["quadparam2"],
            self.nqcCoeffsInput,
            0,
        )

        hlm = {}

        ##55 mode
        modeL = sphtseries.l
        modeM = sphtseries.m
        h55 = sphtseries.mode.data.data  # This is h_55
        hlm[f"{modeL},{modeM}"] = -h55

        ##44 mode
        modeL = sphtseries.next.l
        modeM = sphtseries.next.m
        h44 = sphtseries.next.mode.data.data  # This is h_44
        hlm[f"{modeL},{modeM}"] = -h44

        ##21 mode
        modeL = sphtseries.next.next.l
        modeM = sphtseries.next.next.m
        h21 = sphtseries.next.next.mode.data.data  # This is h_21
        hlm[f"{modeL},{modeM}"] = -h21

        ##33 mode
        modeL = sphtseries.next.next.next.l
        modeM = sphtseries.next.next.next.m
        h33 = sphtseries.next.next.next.mode.data.data  # This is h_33
        hlm[f"{modeL},{modeM}"] = -h33

        ##22 mode
        modeL = sphtseries.next.next.next.next.l
        modeM = sphtseries.next.next.next.next.m
        h22 = sphtseries.next.next.next.next.mode.data.data  # This is h_22
        hlm[f"{modeL},{modeM}"] = -h22

        scaling_factor = self.M * lal.MRSUN_SI / self.distance
        shift = 0
        time_array = (np.arange(0, len(h22)) * self.dt + shift) / (
            self.M * lal.MTSUN_SI
        )
        for key in hlm.keys():
            hlm[key] /= scaling_factor

        self.t = time_array

        self.waveform_modes = hlm
        self.dyn = dyn


class NRHybSur3dq8Model(Model):
    """
    NRHybSur3dq8 waveform

    This model includes the following spin-weighted spherical harmonic modes:
    (2,2), (2,1), (2,0), (3,3), (3,2), (3,1), (3,0), (4,4) (4,3), (4,2) and (5,5)

    """

    def __init__(
        self,
        q: float,
        chi_1: float,
        chi_2: float,
        omega0: float,
        settings: Dict[Any, Any] = None,
    ) -> None:
        """
        Args:
            q (float): Mass ratio m1/m2 >= 1
            chi_1 (float): z component of the dimensionless spin of primary
            chi_2 (float): z component of the dimensionless spin of secondary
            omega0 (float): Initial orbital frequency, in geomtric units
        """

        self.settings = default_NRHybSur3dq8_settings()
        if settings is not None:
            self.settings.update(**settings)

        self.M = self.settings["M"]
        self.dt = self.settings["dt"]
        self.delta_T = self.dt
        if q < 1.0:
            q = 1.0

        self.q = q

        self.chi_1 = chi_1
        self.chi_2 = chi_2
        self.omega0 = omega0

    def __call__(self, sur, parameters=None):
        self._evaluate_model(sur)

    def _evaluate_model(self, sur):

        omega = self.omega0
        f_low = omega / np.pi
        t_sur, h, dyn = sur(
            self.q,
            [0, 0, self.chi_1],
            [0, 0, self.chi_2],
            dt=self.dt,
            f_low=f_low,
            f_ref=f_low,
        )

        self.t = t_sur - t_sur[0]

        self.waveform_modes = {}

        self.waveform_modes["2,2"] = -h[(2, 2)]
        self.waveform_modes["2,1"] = -h[(2, 1)]
        self.waveform_modes["2,0"] = -h[(2, 0)]
        self.waveform_modes["3,3"] = -h[(3, 3)]
        self.waveform_modes["3,2"] = -h[(3, 2)]
        self.waveform_modes["3,1"] = -h[(3, 1)]
        self.waveform_modes["3,0"] = -h[(3, 0)]
        self.waveform_modes["4,4"] = -h[(4, 4)]
        self.waveform_modes["4,3"] = -h[(4, 3)]
        self.waveform_modes["4,2"] = -h[(4, 2)]
        self.waveform_modes["5,5"] = -h[(5, 5)]


class NRHybSur2dq15Model(Model):
    """
       NRHybSur2dq15 waveform

       This model includes the following spin-weighted spherical harmonic modes:
       (2,2), (2,1), (3,3), (4,4) and (5,5)

    |  The surrogate has been trained in the range
    |  q \in [1, 15] and chi1z \in [-0.5, 0.5] chi2z = 0

    """

    def __init__(
        self,
        q: float,
        chi_1: float,
        chi_2: float,
        omega0: float,
        settings: Dict[Any, Any] = None,
    ) -> None:
        """
        Args:
            q (float): Mass ratio m1/m2 >= 1
            chi_1 (float): z component of the dimensionless spin of primary
            chi_2 (float): z component of the dimensionless spin of secondary
            omega0 (float): Initial orbital frequency, in geomtric units
        """

        self.settings = default_NRHybSur2dq15_settings()
        if settings is not None:
            self.settings.update(**settings)

        self.M = self.settings["M"]
        self.dt = self.settings["dt"]
        self.delta_T = self.dt
        if q < 1.0:
            q = 1.0

        self.q = q

        self.chi_1 = chi_1
        self.chi_2 = chi_2
        self.omega0 = omega0

    def __call__(self, sur, parameters=None):
        self._evaluate_model(sur)

    def _evaluate_model(self, sur):

        omega = self.omega0
        f_low = omega / np.pi
        t_sur, h, dyn = sur(
            self.q,
            [0, 0, self.chi_1],
            [0, 0, self.chi_2],
            dt=self.dt,
            f_low=f_low,
            f_ref=f_low,
        )

        self.t = t_sur - t_sur[0]

        self.waveform_modes = {}

        self.waveform_modes["2,2"] = -h[(2, 2)]
        self.waveform_modes["2,1"] = -h[(2, 1)]
        self.waveform_modes["3,3"] = -h[(3, 3)]
        self.waveform_modes["4,4"] = -h[(4, 4)]
        self.waveform_modes["5,5"] = -h[(5, 5)]



import argparse
import importlib
import os,sys
from mimetypes import MimeTypes
import lal
import numpy as np


import matplotlib.pyplot as plt
from pathlib import Path
from pathos.multiprocessing import ProcessingPool as Pool
from pyseobnr.generate_waveform import generate_modes_opt

from pycbc.psd.analytical import aLIGOZeroDetHighPower, aLIGOZeroDetHighPowerGWINC
from pycbc.types import TimeSeries
from pycbc.filter import make_frequency_series
from pycbc.filter.matchedfilter import match
from pycbc.waveform.utils import taper_timeseries

from typing import Dict, Union, Tuple

def combine_modes(
    iota: float, phi: float, modes_dict: Dict
) -> Tuple[np.array, np.array]:

    """Combine modes to compute the waveform polarizations in the direction
    (iota,np.pi/2-phi)

    Args:
        iota (float): Inclination angle (rad)
        phi (float): Azimuthal angle(rad)
        modes_dict (Dict): Dictionary containing the modes, either time of frequency-domain

    Returns:
        np.array: Waveform in the given direction
    """
    sm = 0.0
    for key in modes_dict.keys():
        #print(key)
        ell, m = [int(x) for x in key.split(",")]
        Ylm0 = lal.SpinWeightedSphericalHarmonic(iota, np.pi / 2 - phi, -2, ell, m)
        sm += Ylm0 * modes_dict[key]

    return np.real(sm), -np.imag(sm)


def timeMtoSec(timeM, mt:float):
    return timeM*mt*lal.MTSUN_SI


def ampNRtoPhysicalTD(ampNR, mt:float,distance:float):
    return ampNR*(lal.C_SI*mt*lal.MTSUN_SI)/distance


def generate_v5PHM_waveform(m1:float, m2:float,
                            s1x:float, s1y:float, s1z:float,
                            s2x:float, s2y:float, s2z:float,
                            f_min:float,
                            iota:float, phi:float, distance:float,
                            delta_t:float, approx:str,
                            ell_max: int = 5):

    mtotal = m1+m2
    q = m1/m2
    omega0 = f_min * (np.pi * (m1 + m2) * lal.MTSUN_SI)

    chi_1 = [s1x, s1y, s1z]
    chi_2 = [s2x, s2y, s2z]

    settings = {'ell_max':ell_max,'beta_approx':None,'M':mtotal,"dt":delta_t}
    if approx == 'SEOBNRv5PHM':
        #time, modes, _ = v5P_wrapper.get_EOB_modes(p, ell_max=ell_max)
        time, modes = generate_modes_opt(q,chi_1,chi_2,omega0,approximant=approx,
                           debug=False,settings=settings)
        modes_dict = {}
        for ell in range(2, ell_max + 1):
            for m in range(-ell, ell + 1):
                modes_dict["{},{}".format(ell, m)] = modes["{},{}".format(ell, m)]

    elif approx == 'SEOBNRv5HM':
        time, modes = generate_modes_opt(q,chi_1[-1],chi_2[-1],omega0,approximant=approx,
                                debug=False,settings=settings)

        modes_dict = {}
        for key in modes.keys():
            lm_tag = key.split(',')
            ell,m = lm_tag
            ll = float(ell)
            modes_dict["{},{}".format(ell, m)] = modes[key] # Minus sign to have same convention as v5PHM
            modes_dict["{},-{}".format(ell,m)] = pow(-1.,ll) * np.conjugate(modes[key]) # Minus sign to have same convention as v5PHM


    else:
        raise NotImplementedError



    hp_NR, hc_NR = combine_modes(iota, phi, modes_dict)

    t_s = timeMtoSec(time, mtotal)
    hp = ampNRtoPhysicalTD(hp_NR,mtotal,distance)
    hc = ampNRtoPhysicalTD(hc_NR,mtotal,distance)


    # Taper
    hp_td = TimeSeries(hp, delta_t=delta_t)
    hc_td = TimeSeries(hc, delta_t=delta_t)
    hp_td = taper_timeseries(hp_td, tapermethod="startend")
    hc_td = taper_timeseries(hc_td, tapermethod="startend")

    return time, modes_dict, hp_td, hc_td



def perturbation_mismatch_aligned(m1,m2, s1x,s1y,s1z,s2x,s2y,s2z, iota_s):


    Mt = m1+m2
    q=m1/m2

    if q<1:
        q=1

    phi_s = 0.
    pert=1e-15

    omega0 = 0.02
    f_min = omega0 / (np.pi * (m1 + m2) * lal.MTSUN_SI)
    distance = 1e6*lal.PC_SI
    delta_t = 1./16384.
    approx = 'SEOBNRv5HM'

    # Add small perturbation to m1

    # Generate TD polatizations of SEOBNRv5HM
    time, modes, hp_td, hc_td = generate_v5PHM_waveform(m1, m2,
                                                        s1x, s1y, s1z,
                                                        s2x, s2y, s2z,
                                                        f_min,
                                                        iota_s, phi_s, distance,
                                                        delta_t, approx)
    # Pad zeros
    N = max(len(hp_td), len(hc_td))
    pad = int(2 ** (np.floor(np.log2(N)) + 2))
    hp_td.resize(pad)
    hc_td.resize(pad)

    # Perform the Fourier Transform
    hp = make_frequency_series(hp_td)
    hc = make_frequency_series(hc_td)


    # Generate TD polarizations of the perturbed system
    time_pert, modes_pert, hp_td, hc_td = generate_v5PHM_waveform(m1*(1.+pert), m2,
                                                        s1x, s1y, s1z,
                                                        s2x, s2y, s2z,
                                                        f_min,
                                                        iota_s, phi_s, distance,
                                                        delta_t, approx)

    # Pad zeros
    N = max(len(hp_td), len(hc_td))
    pad = int(2 ** (np.floor(np.log2(N)) + 2))
    hp_td.resize(pad)
    hc_td.resize(pad)

    # Perform the Fourier Transform
    hp_pert = make_frequency_series(hp_td)
    hc_pert = make_frequency_series(hc_td)


    # Generate PSD
    f_low_phys = f_min
    f_high_phys = 2048.
    try:
        psd = aLIGOZeroDetHighPowerGWINC(len(hp), hp.delta_f, f_low_phys)

        # Compute match for hplus
        mm_hp = match(hp,
              hp_pert,
              psd,
              low_frequency_cutoff=f_low_phys,
              high_frequency_cutoff=f_high_phys
              )[0]


        # Compute match for hcross
        mm_hc = match(hc,
              hc_pert,
              psd,
              low_frequency_cutoff=f_low_phys,
              high_frequency_cutoff=f_high_phys
              )[0]

        # Take the mean
        mm_mean  = 1.-np.mean([mm_hp,mm_hc])
        #print(f"mm = {mm_mean}")
    except:

        print(
            f"Error for the following parameters: q = {q}, chi1x = {s1x}, chi1y = {s1y}, chi1z = {s1z}, chi2x = {s2x}, chi2y = {s2y}, chi2z = {s2z}, Mt = {Mt}, iota_s = {iota_s}"
        )
        mm_mean = -1
        pass
    chi1 = [s1x,s1y,s1z]
    chi2 = [s2x,s2y,s2z]

    if mm_mean>0.001:


        plt.figure(figsize = (10,10/1.618))

        mode_list = ['2,2','2,1','3,3','3,2','4,4','4,3','5,5']

        amp22 = np.abs(modes['2,2'])
        idx_max = np.argmax(amp22)
        tmax22 = time[idx_max]

        amp22_pert = np.abs(modes['2,2'])
        idx_max = np.argmax(amp22_pert)
        tmax22_pert = time_pert[idx_max]

        for mode in mode_list:
            plt.plot(time-tmax22, np.real(modes[mode]),label=str(mode))
            plt.plot(time_pert-tmax22_pert, np.real(modes_pert[mode]),ls = '--')


        plt.xlabel("Time (M)")
        plt.ylabel(r"$\|h_{\ell m}|$")
        plt.xlim(-300,100)
        plt.legend(loc=3)
        plt.title(f'MM = {np.round(mm_mean,7)} - q{np.round(q,3)}_s{np.round(chi1,3)}_s{np.round(chi2,3)} - iota = {np.round(iota_s,3)}')
        Path('./plots').mkdir(parents=True, exist_ok=True)
        plt.savefig(f'./plots/pert_q{np.round(q,3)}_s{np.round(chi1,3)}_s{np.round(chi2,3)}_i{np.round(iota_s,3)}.png', dpi = 200)
        plt.clf()

    res = [m1,m2, *chi1,*chi2, mm_mean,iota_s]

    return (m1,m2, *chi1,*chi2,iota_s, mm_mean)



def process_one_case(input):
    m1,m2,  s1x,s1y,s1z,s2x,s2y,s2z, iota_s = input

    m1,m2,  s1x,s1y,s1z,s2x,s2y,s2z, iota_s, mm = perturbation_mismatch_aligned(m1,m2, s1x,s1y,s1z,s2x,s2y,s2z, iota_s)
    return np.array([m1,m2,s1x,s1y,s1z,s2x,s2y,s2z, iota_s, mm])


if __name__ == "__main__":

    p = argparse.ArgumentParser(
        description="Perturbation test for the full polarizations"
    )
    p.add_argument(
        "--points", type=int, help="Number of points", default="10000"
    )
    p.add_argument(
        "--plots", action="store_true", help="Make diagnostic plots", default=True
    )
    p.add_argument(
        "--ncores", type=int, help="Number of cores to use", default="64"
    )
    p.add_argument(
        "--q-max", type=float, help="Maximum mass-ratio", default="100.0"
    )
    p.add_argument(
        "--M-min", type=float, help="Minimum total mass", default="10.0"
    )
    p.add_argument(
        "--M-max", type=float, help="Maximum total mass", default="300.0"
    )
    p.add_argument(
        "--a-max", type=float, help="Maximum spin magnitude", default="0.995"
    )

    args = p.parse_args()

    seed=150914
    N = args.points
    a_max = args.a_max

    np.random.seed(seed)
    q = np.random.uniform(1., args.q_max, N)
    #m1 = np.random.uniform(args.M_min,args.M_max,N)
    np.random.seed(seed-1)
    mtotal = np.random.uniform(args.M_min, args.M_max, N)

    m1 = q/(1.+q)*mtotal
    m2 = 1./(1.+q)*mtotal

    np.random.seed(seed+999998)
    chi1x = np.zeros_like(q)
    chi1y = np.zeros_like(q)
    chi1z = np.random.uniform(-a_max,a_max,N)

    np.random.seed(seed+999997)
    chi2x = np.zeros_like(q)
    chi2y = np.zeros_like(q)
    chi2z = np.random.uniform(-a_max,a_max,N)

    np.random.seed(seed+999996)
    cos_iota = np.random.uniform(-1.0, 1.0, N)
    iota_arr = np.arccos(cos_iota)

    lst = [(a,b,c,d,e,f,g,h,i) for a,b,c,d,e,f,g,h,i in zip(m1,m2,chi1x,chi1y,chi1z,chi2x,chi2y,chi2z,iota_arr)]

    with Pool(args.ncores) as pool:
        all_means = pool.map(process_one_case, lst)

    all_means = np.array(all_means)

    np.savetxt(f"perturbation_mismatch.dat", all_means)

    ### Plots ###

    if args.plots:

        import matplotlib
        import matplotlib.pyplot as plt

        plt_dir = './plots'
        Path(plt_dir).mkdir(parents=True, exist_ok=True)

        res_path = f"perturbation_mismatch.dat"
        res = np.loadtxt(res_path)

        m1 = res[:,0]
        m2 = res[:,1]
        q = m1/m2

        chi1x = res[:,2]
        chi1y = res[:,3]
        chi1z = res[:,4]

        chi2x = res[:,5]
        chi2y = res[:,6]
        chi2z = res[:,7]

        iota = res[:,8]
        mm = res[:,9]

        nu = q/(1+q)**2
        m1 = q/(1+q)
        m2 = 1/(1 + q)
        m_total = m1 + m2
        ap = (m1*chi1z+m2*chi2z)/m_total
        am = (m1*chi1z-m2*chi2z)/m_total

        # Mismatch across parameter space
        mm_s, q_s, ap_s, am_s = map(list, zip(*sorted(zip(mm, q, ap, am))))
        plt.scatter(q_s, ap_s, c=mm_s,linewidths=1,norm=matplotlib.colors.LogNorm())
        plt.xlabel('$q$')
        plt.ylabel('$a_{+}$')
        plt.title('$\mathcal{M}_{\mathrm{median}} = $' + f'{np.round(np.median(mm), 7)}')
        cbar=plt.colorbar()
        cbar.set_label('$\mathcal{M}$')
        plt.savefig(f'{plt_dir}/mm_pert_apam.png', bbox_inches = 'tight', dpi = 300)
        plt.close()

        # Histogram
        plt.hist(
            mm,
            bins=np.logspace(start=np.log10(1e-16), stop=np.log10(1.0), num=50),
            alpha=0.4,
        )
        plt.axvline(np.median(mm), c="C0", ls="--")
        plt.gca().set_xscale("log")
        plt.xlabel("$\mathcal{M}$")
        plt.title(
            "$\mathcal{M}_{\mathrm{median}} = $" +
            f"{np.round(np.median(mm),12)}" +
            "$\mathcal{M}_{\mathrm{max}} = $" +
            f"{np.round(np.max(mm),6)}"
        )
        plt.savefig(f'{plt_dir}/mm_hist.png', bbox_inches = 'tight', dpi = 300)
        plt.close()



"""
Additional utility functions.
"""

from typing import Any, Dict

import numpy as np


def frame_inv_amp(modes: Dict[Any, Any], ell_max: int = 2, use_symm=True) -> np.ndarray:
    """Compute the frame-invariant amplitude.
    By default, assumes that we have aligned-spin symmetry.

    Args:
        modes (Dict[Any,Any]): The dictionary of modes
        ell_max (int): The maximum l to use
        use_symm (bool): Assume up-down aligned symmetry

    Returns:
        np.ndarray: The frame-invariant amplitude time series
    """
    total = 0.0

    for mode in modes.keys():
        ell, m = mode
        if ell > ell_max:
            continue
        if use_symm:
            total += 2 * np.abs(modes[mode]) ** 2
        else:
            total += np.abs(modes[mode]) ** 2
    return np.sqrt(total)



# cython: language_level=3, boundscheck=False, cdivision=True, wraparound=False,
# cython: profile=True, linetrace=True, binding=True, initializedcheck=False
"""
Contains the actual RHS of the EOM, wrapped in cython.
This allows some of the cython functions used in the RHS to be called more efficiently.
"""

import numpy as np
cimport numpy as np

from pyseobnr.eob.utils.containers cimport EOBParams
from pyseobnr.eob.waveform.waveform cimport RadiationReactionForce
from pyseobnr.eob.hamiltonian.Hamiltonian_C cimport Hamiltonian_C


cpdef get_rhs(
    double t,
    double[::1] z,
    Hamiltonian_C H,
    RadiationReactionForce RR,
    double chi_1,
    double chi_2,
    double m_1,
    double m_2,
    EOBParams params
):
    """
    Compute the RHS of the EOB evolution equations.
    In particular this function returns

    * :math:`\\dot{r}`,
    * :math:`\\dot{\\phi}`,
    * :math:`\\dot{p}_{r}`,
    * and :math:`\\dot{p}_{\\phi}`.

    See for example Eq(2) of [Buades2021]_ .

    The Hamiltonian is given by Eq(14) in [SEOBNRv5HM]_ doc
    and explicitly spelled out in Section I.C of [SEOBNRv5HM-theory]_
    and the RR force is described in Eq(43) of [SEOBNRv5HM]_ document, both
    contained in [DCC_T2300060]_ .

    """
    cdef double[::1] q = z[:2]
    cdef double[::1] p = z[2:]

    cdef double H_val, omega, omega_circ, xi
    cdef double dynamics[6]
    dynamics[:] = H.dynamics(q, p, chi_1, chi_2, m_1, m_2)
    H_val = dynamics[4]
    omega = dynamics[3]
    params.dynamics.p_circ[1] = p[1]

    omega_circ = H.omega(q, params.dynamics.p_circ, chi_1, chi_2, m_1, m_2)

    xi = dynamics[5]

    cdef (double, double) RR_f = RR.RR(q, p, omega, omega_circ, H_val, params)

    cdef double deriv[4]
    deriv[:] = [xi * dynamics[2], dynamics[3], -dynamics[0] * xi + RR_f[0], -dynamics[1] + RR_f[1]]
    return deriv


cpdef augment_dynamics(
    double[:, :] dynamics,
    double chi_1,
    double chi_2,
    double m_1,
    double m_2,
    Hamiltonian_C H
):
    """Compute dynamical quantities we need for the waveform

    Args:
        dynamics (np,ndarray): The dynamics array: t,r,phi,pr,pphi
    """
    result = []
    cdef double[:] p_c = np.zeros(2)
    cdef double H_val, omega, omega_c
    cdef double dyn[6]
    cdef double[:] q
    cdef double[:] p
    cdef double[:] row
    cdef int i, N
    N = dynamics.shape[0]

    for i in range(N):
        row = dynamics[i]
        q = row[1:3]
        p = row[3:5]
        p_c[1] = p[1]

        # Evaluate a few things: H, omega, omega_circ

        dyn[:] = H.dynamics(q, p, chi_1, chi_2, m_1, m_2)
        omega = dyn[3]
        H_val = dyn[4]

        omega_c = H.omega(q, p_c, chi_1, chi_2, m_1, m_2)

        result.append([H_val, omega, omega_c])
    result = np.array(result)
    return np.c_[dynamics, result]



from typing import get_args
from unittest import mock

import lal
import numpy as np

import pytest

import pyseobnr
from pyseobnr.generate_waveform import (
    GenerateWaveform,
    SupportedApproximants,
    generate_modes_opt,
)


@pytest.fixture
def basic_settings():
    m1 = 50.0
    m2 = 30.0
    params_dict = {
        "mass1": m1,
        "mass2": m2,
        "spin1x": 0,
        "spin1y": 0,
        "spin1z": 0.5,
        "spin2x": 0,
        "spin2y": 0,
        "spin2z": 0.1,
        "deltaT": 1 / 2048.0,
        "deltaF": 0.125,
        "f22_start": 0.0157 / ((m1 + m2) * np.pi * lal.MTSUN_SI),
        "phi_ref": 0.0,
        "distance": 1.0,
        "inclination": np.pi / 3.0,
        "f_max": 1024.0,
        "approximant": "SEOBNRv5HM",
        "postadiabatic": False,
    }
    return params_dict


def test_mode_arrays_settings(basic_settings):
    """Checks the behaviour wrt. mode_array setting"""

    # cannot accept both
    value_error = (
        r"Only one setting can be specified between .?ModeArray.? and .?mode_array.?."
    )

    with pytest.raises(
        ValueError,
        match=value_error,
    ):
        _ = GenerateWaveform(basic_settings | {"mode_array": [], "ModeArray": []})

    with pytest.raises(
        ValueError,
        match=value_error,
    ):
        _ = GenerateWaveform(
            basic_settings
            | {"approximant": "SEOBNRv5PHM", "mode_array": [], "ModeArray": []}
        )

    calls_to_check = (
        "generate_td_modes",
        "generate_td_polarizations",
        "generate_fd_polarizations",
    )

    # works with list and tuples
    for approximant in "SEOBNRv5HM", "SEOBNRv5PHM":
        wfm_gen = GenerateWaveform(
            basic_settings
            | {"approximant": approximant, "mode_array": [(2, 2), (3, 3)]}
        )

        for func in calls_to_check:
            _ = getattr(wfm_gen, func)()

        wfm_gen = GenerateWaveform(
            basic_settings
            | {"approximant": approximant, "mode_array": ((2, 2), (3, 3))}
        )

        for func in calls_to_check:
            _ = getattr(wfm_gen, func)()

        # incorrect mode array yields an error
        wfm_gen = GenerateWaveform(
            basic_settings | {"approximant": approximant, "mode_array": [2, 2, 3, 3]}
        )
        for func in calls_to_check:
            with pytest.raises(TypeError):
                _ = getattr(wfm_gen, func)()


def test_f_ref_behaviour(basic_settings):
    """When not set, f_ref should be set to f22_start"""

    assert "f_ref" not in basic_settings  # for this test to work

    approximants = "SEOBNRv5HM", "SEOBNRv5PHM"

    for approximant in approximants:
        wfm_gen = GenerateWaveform(
            basic_settings | {"approximant": approximant, "f_ref": 21, "f22_start": 20}
        )

        assert wfm_gen.parameters["f_ref"] == 21

    # takes value of f22_start if not specified
    for approximant in approximants:
        wfm_gen = GenerateWaveform(
            basic_settings | {"approximant": approximant, "f22_start": 21}
        )

        assert wfm_gen.parameters["f_ref"] == 21

    # inherit default value
    for approximant in approximants:
        new_settings = basic_settings | {"approximant": approximant}
        new_settings.pop("f22_start")
        wfm_gen = GenerateWaveform(new_settings)

        assert wfm_gen.parameters["f_ref"] == wfm_gen.parameters["f22_start"]

    # sanity check
    value_error = r"f_ref has to be positive!"
    with pytest.raises(
        ValueError,
        match=value_error,
    ):
        _ = GenerateWaveform(basic_settings | {"f_ref": -1})


def test_generate_modes_opt_settings_all_models(basic_settings):
    """Checks error reporting of incorrect parameters to generate_modes_opt"""

    with pytest.raises(
        ValueError,
        match="mass-ratio has to be positive and with convention q>=1",
    ):
        _ = generate_modes_opt(
            q=0.9,
            chi1=0,
            chi2=0,
            omega_start=0,
        )

    with pytest.raises(
        ValueError,
        match="omega_start has to be positive",
    ):
        _ = generate_modes_opt(
            q=1.1,
            chi1=0,
            chi2=0,
            omega_start=-0.00001,
        )


def test_generate_modes_opt_settings_hm(basic_settings):
    """Checks error reporting of incorrect parameters to generate_modes_opt for HM"""

    # pytest-mocker is hard to use with context managers
    with mock.patch.object(
        pyseobnr.generate_waveform.SEOBNRv5HM.SEOBNRv5HM_opt,
        "__call__",
        autospec=True,
    ) as p_model_call:
        with pytest.raises(
            ValueError,
            match="Dimensionless spin magnitudes cannot be greater than 1!",
        ):
            _ = generate_modes_opt(
                q=1.1,
                chi1=1.1,
                chi2=0,
                omega_start=0.00001,
                approximant="SEOBNRv5HM",
            )
            p_model_call.assert_not_called()

    with mock.patch.object(
        pyseobnr.generate_waveform.SEOBNRv5HM.SEOBNRv5HM_opt,
        "__call__",
        autospec=True,
    ) as p_model_call:

        def compute_dynamics(self):
            self.t = "something"
            self.waveform_modes = "something else"

        p_model_call.side_effect = compute_dynamics

        _ = generate_modes_opt(
            q=1.1,
            chi1=1,
            chi2=0,
            omega_start=0.00001,
            approximant="SEOBNRv5HM",
        )
        p_model_call.assert_called_once()

    with mock.patch.object(
        pyseobnr.generate_waveform.SEOBNRv5HM.SEOBNRv5HM_opt,
        "__call__",
        autospec=True,
    ) as p_model_call:

        def compute_dynamics(self):
            self.t = "something"
            self.waveform_modes = "something else"

        p_model_call.side_effect = compute_dynamics

        _ = generate_modes_opt(
            q=1.1,
            chi1=1,
            chi2=0,
            omega_start=0.00001,
            approximant="SEOBNRv5HM",
        )
        p_model_call.assert_called_once()


def test_generate_modes_opt_precessing_chi_array_float_int():
    """models should accept chi arrays with only z component, ints and floats"""
    q = 41.83615272380585
    omega0 = 0.02

    for chi_2 in (0, 0.3):
        chi_1 = np.array([0.0, 0.0, 0.98917404])

        approx: SupportedApproximants
        for approx in get_args(SupportedApproximants):
            _ = generate_modes_opt(
                q,
                chi_1,
                chi_2,
                omega0,
                approximant=approx,
                debug=False,
                settings={
                    "ell_max": 5,
                    "beta_approx": None,
                    "M": 154.2059835575123,
                    "dt": 6.103515625e-05,
                },
            )

    for chi_1 in (0, 0.3):
        chi_2 = np.array([0.0, 0.0, 0.98917404])

        approx: SupportedApproximants
        for approx in get_args(SupportedApproximants):
            _ = generate_modes_opt(
                q,
                chi_1,
                chi_2,
                omega0,
                approximant=approx,
                debug=False,
                settings={
                    "ell_max": 5,
                    "beta_approx": None,
                    "M": 154.2059835575123,
                    "dt": 6.103515625e-05,
                },
            )



# cython: language_level=3
from pyseobnr.eob.utils.containers cimport EOBParams, FluxParams
cdef extern from "eob_parameters.h":
    const int PN_limit
    const int ell_max

cpdef (double, double) RR_force(
    double[::1] q,
    double[::1] p,
    double omega,
    double omega_circ,
    double H,
    EOBParams eob_pars)

cdef class RadiationReactionForce:
    cpdef (double, double) RR(
        self,
        double[::1] q,
        double[::1] p,
        double omega,
        double omega_circ,
        double H,
        EOBParams eob_pars)

cdef class SEOBNRv5RRForce(RadiationReactionForce):
    cpdef (double, double) RR(
        self,
        double[::1] q,
        double[::1] p,
        double omega,
        double omega_circ,
        double H,
        EOBParams eob_par)



import argparse
import importlib
import os
import sys

import gwsurrogate
import numpy as np
from bilby.core.utils import check_directory_exists_and_if_not_mkdir
from scipy.interpolate import InterpolatedUnivariateSpline
from pathos.multiprocessing import ProcessingPool as Pool

sys.path.append(
    os.path.join(os.path.abspath(os.path.dirname(__file__)), "../../calibration")
)
sys.path.append(
    os.path.join(os.path.abspath(os.path.dirname(__file__)), "../../MR_pipeline")
)
sys.path.append(
    os.path.join(
        os.path.abspath(os.path.dirname(__file__)), "../../toys/hamiltonian_prototype"
    )
)
from metrics import *
from models import *
from IV_fits import *

from parameters import parameters_random, parameters_random_2D


mode_list_v5HM = [(2, 2), (2, 1), (3, 3), (3, 2), (4, 4), (4, 3), (5, 5)]
mode_list_v4HM = [(2, 2), (2, 1), (3, 3), (4, 4), (5, 5)]


sur_3dq8 = gwsurrogate.LoadSurrogate("NRHybSur3dq8")
sur_2dq15 = gwsurrogate.LoadSurrogate("NRHybSur2dq15")


def NRSur_IV(modes, t, ell, m):
    if (
        (ell == 2 and m == 2)
        or (ell == 3 and m == 3)
        or (ell == 2 and m == 1)
        or (ell == 4 and m == 4)
        or (ell == 3 and m == 2)
        or (ell == 4 and m == 3)
    ):
        tpeak = t[np.argmax(abs(modes["2,2"]))]
    elif ell == 5 and m == 5:
        tpeak = t[np.argmax(abs(modes["2,2"]))] - 10
    else:
        return NotImplementedError

    hlm_spline = InterpolatedUnivariateSpline(t, abs(modes[f"{ell},{m}"]))
    hlmdot_spline = hlm_spline.derivative()
    hlmdotdot_spline = hlm_spline.derivative(2)
    philm_spline = InterpolatedUnivariateSpline(
        t, (np.unwrap(np.angle(modes[f"{ell},{m}"])))
    )
    omegalm_spline = philm_spline.derivative()
    omegalmdot_spline = philm_spline.derivative(2)

    hlm = hlm_spline(tpeak)
    hlm_dot = hlmdot_spline(tpeak)
    hlm_dotdot = hlmdotdot_spline(tpeak)
    omegalm = omegalm_spline(tpeak)
    omegalm_dot = omegalmdot_spline(tpeak)

    return hlm, hlm_dot, hlm_dotdot, omegalm, omegalm_dot


def IV_difference_NRHybSur(
    q: float,
    chi1: float,
    chi2: float,
    model_name: str,
):
    if model_name == "NRHybSur2dq15":
        target_model = NRHybSur2dq15Model(
            q,
            chi1,
            chi2,
            0.015,
        )
        sur = sur_2dq15
        mode_list = mode_list_v4HM
    elif model_name == "NRHybSur3dq8":
        target_model = NRHybSur3dq8Model(
            q,
            chi1,
            chi2,
            0.015,
        )
        sur = sur_3dq8
        mode_list = mode_list_v5HM
    target_model(sur)

    m1 = q / (1 + q)
    m2 = 1 - m1
    nu = q/(1 + q)**2
    fits_IV = InputValueFits(m1, m2, [0.0, 0.0, chi1], [0.0, 0.0, chi2])

    hlm_diff = []
    hlm_dot_diff = []
    hlm_dotdot_diff = []
    omegalm_diff = []
    omegalm_dot_diff = []
    for mode in mode_list:
        ell, m = mode
        hlm, hlm_dot, hlm_dotdot, omegalm, omegalm_dot = NRSur_IV(
            target_model.waveform_modes, target_model.t, ell=ell, m=m
        )
        hlm_diff.append(hlm/nu - fits_IV.habs()[ell, m])
        hlm_dot_diff.append(hlm_dot/nu - fits_IV.hdot()[ell, m])
        hlm_dotdot_diff.append(hlm_dotdot/nu - fits_IV.hdotdot()[ell, m])
        omegalm_diff.append(omegalm - fits_IV.omega()[ell, m])
        omegalm_dot_diff.append(omegalm_dot - fits_IV.omegadot()[ell, m])

    return (
        target_model.q,
        target_model.chi_1,
        target_model.chi_2,
        np.array(hlm_diff),
        np.array(hlm_dot_diff),
        np.array(hlm_dotdot_diff),
        np.array(omegalm_diff),
        np.array(omegalm_dot_diff),
    )


def process_one_case(input):
    q, chi1, chi2, model_name = input
    (
        q,
        chi1,
        chi2,
        hlm_diff,
        hlm_dot_diff,
        hlm_dotdot_diff,
        omegalm_diff,
        omegalm_dot_diff,
    ) = IV_difference_NRHybSur(q, chi1, chi2, model_name)
    return hlm_diff, hlm_dot_diff, hlm_dotdot_diff, omegalm_diff, omegalm_dot_diff


if __name__ == "__main__":
    p = argparse.ArgumentParser(
        description="Compute input values difference against NRSur"
    )
    p.add_argument("--points", type=int, help="Number of points", default="5000")
    p.add_argument("--chi-max", type=float, help="Maximum spin", default="0.8")
    p.add_argument("--q-max", type=float, help="Maximum mass-ratio", default="8.0")
    p.add_argument(
        "--name", type=str, help="Name of the output file", default="iv_difference"
    )
    p.add_argument(
        "--plots", action="store_true", help="Make diagnostic plots", default=True
    )
    p.add_argument(
        "--model-name",
        type=str,
        help="Name of the surrogate model",
        default="NRHybSur3dq8",
    )
    p.add_argument(
        "--plots-only",
        action="store_true",
        help="Only generate plots, don't recompute things",
    )
    p.add_argument("--n-cpu", type=int, help="Number of cores to use", default=64)
    p.add_argument(
        "--include-all",
        action="store_true",
        help="Include odd m modes close to equal mass",
        default=False,
    )
    args = p.parse_args()

    if args.model_name == "NRHybSur2dq15":
        mode_list = mode_list_v4HM
    elif args.model_name == "NRHybSur3dq8":
        mode_list = mode_list_v5HM

    if not args.plots_only:

        if args.model_name == "NRHybSur2dq15":
            qarr, chi1arr = parameters_random_2D(
                args.points, 1.0, args.q_max, -args.chi_max, args.chi_max
            )
            lst = [(q, chi1, 0.0, args.model_name) for q, chi1 in zip(qarr, chi1arr)]
            chi2arr = np.zeros(args.points)
        elif args.model_name == "NRHybSur3dq8":
            qarr, chi1arr, chi2arr = parameters_random(
                args.points,
                1.0,
                args.q_max,
                -args.chi_max,
                args.chi_max,
                -args.chi_max,
                args.chi_max,
            )
            lst = [
                (q, chi1, chi2, args.model_name)
                for q, chi1, chi2 in zip(qarr, chi1arr, chi2arr)
            ]

        pool = Pool(args.n_cpu)
        all_means = pool.map(process_one_case, lst)

        all_means = np.array(all_means)
        # This is an array with shape n cases x m quantities x p modes
        hlm_diffs = {}
        hlm_dot_diffs = {}
        hlm_dotdot_diffs = {}
        omegalm_diffs = {}
        omegalm_dot_diffs = {}

        for mode in mode_list:
            hlm_diffs[mode] = []
            hlm_dot_diffs[mode] = []
            hlm_dotdot_diffs[mode] = []
            omegalm_diffs[mode] = []
            omegalm_dot_diffs[mode] = []
        for i in range(len(all_means)):
            [
                hlm_for_case,
                hlmdot_for_case,
                hlmddot_for_case,
                omega_for_case,
                omegadot_for_case,
            ] = all_means[i]
            for j, mode in enumerate(mode_list):
                hlm_diffs[mode].append(hlm_for_case[j])
                hlm_dot_diffs[mode].append(hlmdot_for_case[j])
                hlm_dotdot_diffs[mode].append(hlmddot_for_case[j])
                omegalm_diffs[mode].append(omega_for_case[j])
                omegalm_dot_diffs[mode].append(omegadot_for_case[j])

        for mode in mode_list:
            ell, m = mode
            np.savetxt(f"{args.name}_{args.model_name}_h{ell}{m}.dat", hlm_diffs[mode])
            np.savetxt(
                f"{args.name}_{args.model_name}_hdot{ell}{m}.dat", hlm_dot_diffs[mode]
            )
            np.savetxt(
                f"{args.name}_{args.model_name}_hddot{ell}{m}.dat",
                hlm_dotdot_diffs[mode],
            )
            np.savetxt(
                f"{args.name}_{args.model_name}_omega{ell}{m}.dat", omegalm_diffs[mode]
            )
            np.savetxt(
                f"{args.name}_{args.model_name}_omegadot{ell}{m}.dat",
                omegalm_dot_diffs[mode],
            )

        np.savetxt(f"parameters_{args.model_name}.dat", np.c_[qarr, chi1arr, chi2arr])

    ### Plots ###

    if args.plots:

        import matplotlib
        import matplotlib.pyplot as plt

        plt_dir = "./plots"
        check_directory_exists_and_if_not_mkdir(plt_dir)

        res_path = args.name
        params = np.genfromtxt(f"parameters_{args.model_name}.dat")
        q = params[:, 0]
        chi1 = params[:, 1]
        chi2 = params[:, 2]

        nu = q / (1 + q) ** 2
        m1 = q / (1 + q)
        m2 = 1 / (1 + q)
        ap = m1 * chi1 + m2 * chi2
        am = m1 * chi1 - m2 * chi2
        for mode in mode_list:
            ell, m = mode
            h_diff = np.loadtxt(f"{args.name}_{args.model_name}_h{ell}{m}.dat")
            hd_diff = np.loadtxt(f"{args.name}_{args.model_name}_hdot{ell}{m}.dat")
            hdd_diff = np.loadtxt(f"{args.name}_{args.model_name}_hddot{ell}{m}.dat")
            om_diff = np.loadtxt(f"{args.name}_{args.model_name}_omega{ell}{m}.dat")
            omd_diff = np.loadtxt(f"{args.name}_{args.model_name}_omegadot{ell}{m}.dat")

            if not args.include_all:
                if m % 2 == 1:
                    idx_notEMES = np.where((q > 1.01) | (np.abs(chi1 - chi2) > 0.01))[0]
                    q_pl = q[idx_notEMES]
                    ap_pl = ap[idx_notEMES]
                    h_diff = h_diff[idx_notEMES]
                    hd_diff = hd_diff[idx_notEMES]
                    hdd_diff = hdd_diff[idx_notEMES]
                    om_diff = om_diff[idx_notEMES]
                    omd_diff = omd_diff[idx_notEMES]

                else:
                    q_pl = q
                    ap_pl = ap
            else:
                q_pl = q
                ap_pl = ap

            # |h|

            # Histogram
            plt.hist(
                h_diff,
                bins=np.logspace(start=np.log10(0.00001), stop=np.log10(1.0), num=50),
                alpha=0.4,
                label=args.model_name + " - SEOBNRV5HM",
            )
            rmse = np.sqrt(np.sum(h_diff ** 2) / len(h_diff))
            plt.legend(loc="best")
            plt.gca().set_xscale("log")
            plt.xlabel(f"$\Delta |h{ell}{m}|$")
            plt.title("RMSE = " + f"{np.round(rmse,6)}")
            plt.savefig(
                f"{plt_dir}/hist_{args.model_name}_h{ell}{m}.png",
                bbox_inches="tight",
                dpi=300,
            )
            plt.close()

            # Difference across parameter space
            h_diff_s, q_s, ap_s = map(list, zip(*sorted(zip(h_diff, q_pl, ap_pl))))
            plt.scatter(
                q_s,
                ap_s,
                c=h_diff_s,
                linewidths=1,
                # norm=matplotlib.colors.LogNorm()
            )
            plt.ylabel("$\chi_{\mathrm{eff}}$")
            plt.xlabel("$q$")
            cbar = plt.colorbar()
            cbar.set_label(f"$\Delta |h{ell}{m}|$")
            plt.savefig(
                f"{plt_dir}/scatter_{args.model_name}_h{ell}{m}.png",
                bbox_inches="tight",
                dpi=300,
            )
            plt.close()

            # |\dot h|

            # Histogram
            plt.hist(
                hd_diff,
                bins=np.logspace(start=np.log10(0.00001), stop=np.log10(1.0), num=50),
                alpha=0.4,
                label=args.model_name + " - SEOBNRV5HM",
            )
            rmse = np.sqrt(np.sum(hd_diff ** 2) / len(hd_diff))
            plt.legend(loc="best")
            plt.gca().set_xscale("log")
            plt.xlabel(f"$\Delta d|h{ell}{m}|/dt$")
            plt.title("RMSE = " + f"{np.round(rmse,6)}")
            plt.savefig(
                f"{plt_dir}/hist_{args.model_name}_hdot{ell}{m}.png",
                bbox_inches="tight",
                dpi=300,
            )
            plt.close()

            # Difference across parameter space
            hd_diff_s, q_s, ap_s = map(list, zip(*sorted(zip(hd_diff, q_pl, ap_pl))))
            plt.scatter(
                q_s,
                ap_s,
                c=hd_diff_s,
                linewidths=1,
                # norm=matplotlib.colors.LogNorm()
            )
            plt.ylabel("$\chi_{\mathrm{eff}}$")
            plt.xlabel("$q$")
            cbar = plt.colorbar()
            cbar.set_label(f"$\Delta d|h{ell}{m}|/dt$")
            plt.savefig(
                f"{plt_dir}/scatter_{args.model_name}_hdot{ell}{m}.png",
                bbox_inches="tight",
                dpi=300,
            )
            plt.close()

            # |\ddot h|

            # Histogram
            plt.hist(
                hdd_diff,
                bins=np.logspace(start=np.log10(0.00001), stop=np.log10(1.0), num=50),
                alpha=0.4,
                label=args.model_name + " - SEOBNRV5HM",
            )
            rmse = np.sqrt(np.sum(hdd_diff ** 2) / len(hdd_diff))
            plt.legend(loc="best")
            plt.gca().set_xscale("log")
            plt.xlabel(f"$\Delta d^2|h{ell}{m}|/dt^2$")
            plt.title("RMSE = " + f"{np.round(rmse,6)}")
            plt.savefig(
                f"{plt_dir}/hist_{args.model_name}_hddot{ell}{m}.png",
                bbox_inches="tight",
                dpi=300,
            )
            plt.close()

            # Difference across parameter space
            hdd_diff_s, q_s, ap_s = map(list, zip(*sorted(zip(hdd_diff, q_pl, ap_pl))))
            plt.scatter(
                q_s,
                ap_s,
                c=hdd_diff_s,
                linewidths=1,
                # norm=matplotlib.colors.LogNorm()
            )
            plt.ylabel("$\chi_{\mathrm{eff}}$")
            plt.xlabel("$q$")
            cbar = plt.colorbar()
            cbar.set_label(f"$\Delta d^2|h{ell}{m}|/dt^2$")
            plt.savefig(
                f"{plt_dir}/scatter_{args.model_name}_hddot{ell}{m}.png",
                bbox_inches="tight",
                dpi=300,
            )
            plt.close()

            # \omega

            # Histogram
            plt.hist(
                om_diff,
                bins=np.logspace(start=np.log10(0.00001), stop=np.log10(1.0), num=50),
                alpha=0.4,
                label=args.model_name + " - SEOBNRV5HM",
            )
            rmse = np.sqrt(np.sum(om_diff ** 2) / len(om_diff))
            plt.legend(loc="best")
            plt.gca().set_xscale("log")
            plt.xlabel(f"$\Delta \omega{ell}{m}$")
            plt.title("RMSE = " + f"{np.round(rmse,6)}")
            plt.savefig(
                f"{plt_dir}/hist_{args.model_name}_omega{ell}{m}.png",
                bbox_inches="tight",
                dpi=300,
            )
            plt.close()

            # Difference across parameter space
            om_diff_s, q_s, ap_s = map(list, zip(*sorted(zip(om_diff, q_pl, ap_pl))))
            plt.scatter(
                q_s,
                ap_s,
                c=om_diff_s,
                linewidths=1,
                # norm=matplotlib.colors.LogNorm()
            )
            plt.ylabel("$\chi_{\mathrm{eff}}$")
            plt.xlabel("$q$")
            cbar = plt.colorbar()
            cbar.set_label(f"$\Delta \omega{ell}{m}$")
            plt.savefig(
                f"{plt_dir}/scatter_{args.model_name}_omega{ell}{m}.png",
                bbox_inches="tight",
                dpi=300,
            )
            plt.close()

            # \dot \omega

            # Histogram
            plt.hist(
                omd_diff,
                bins=np.logspace(start=np.log10(0.00001), stop=np.log10(1.0), num=50),
                alpha=0.4,
                label=args.model_name + " - SEOBNRV5HM",
            )
            rmse = np.sqrt(np.sum(omd_diff ** 2) / len(omd_diff))
            plt.legend(loc="best")
            plt.gca().set_xscale("log")
            plt.xlabel(f"$\Delta d\omega{ell}{m}/dt$")
            plt.title("RMSE = " + f"{np.round(rmse,6)}")
            plt.savefig(
                f"{plt_dir}/hist_{args.model_name}_omegadot{ell}{m}.png",
                bbox_inches="tight",
                dpi=300,
            )
            plt.close()

            # Difference across parameter space
            omd_diff_s, q_s, ap_s = map(list, zip(*sorted(zip(omd_diff, q_pl, ap_pl))))
            plt.scatter(
                q_s,
                ap_s,
                c=omd_diff_s,
                linewidths=1,
                # norm=matplotlib.colors.LogNorm()
            )
            plt.ylabel("$\chi_{\mathrm{eff}}$")
            plt.xlabel("$q$")
            cbar = plt.colorbar()
            cbar.set_label(f"$\Delta d\omega{ell}{m}/dt$")
            plt.savefig(
                f"{plt_dir}/scatter_{args.model_name}_omegadot{ell}{m}.png",
                bbox_inches="tight",
                dpi=300,
            )
            plt.close()



#!/usr/bin/env python3
"""
Computes post-adiabatic initial conditions in polar coordinates.
"""

from typing import Callable

import numpy as np

from ..fits.fits_Hamiltonian import dSO as dSO_poly_fit
from ..hamiltonian import Hamiltonian
from .initial_conditions_aligned_precessing import computeIC_augm
from .postadiabatic_C_prec import compute_postadiabatic_dynamics


def compute_IC_PA(
    omega_ref: float,
    omega_start: float,
    H: Hamiltonian,
    RR: Callable,
    chi1_v: np.ndarray,
    chi2_v: np.ndarray,
    m_1: float,
    m_2: float,
    splines: dict,
    t_pn: np.array,
    dynamics_pn: np.array,
    **kwargs,
):
    """Compute the postadiabatic initial conditions for a precessing-spin BBH binary

    Args:
        omega_ref (float): Reference *orbital* frequency in geometric units at which the spins are defined
        omega_start (float): Initial *orbital* frequency in geometric units
        H (Hamiltonian): The Hamiltonian object to use
        RR (function): The RR force to use
        chi1_v (np.ndarray): Dimensionless spin vector of the primary
        chi2_v (np.ndarray): Dimensionless spin vector of the secondary
        m_1 (float): Mass of primary
        m_2 (float): Mass of secondary
        splines (dict): Dictionary containing the splines in orbital frequency of the vector components
                        of the spins, LN and L as well as the spin projections onto LN and L
        t_pn (np.array): Time array of the PN evolution of the spins and Newtonian angular momentum.
        dynamics_pn (np.array): Array of the spin-precessing PN evolution. It contains the Newtonian
                        angular momentum, the dimensionful spin vectors and the PN orbital frequency.

    Returns:
        tuple: The initial conditions: (r,pphi,pr)
    """

    params = kwargs["params"]
    postadiabatic_type = kwargs["postadiabatic_type"]
    PA_success = False

    try:
        postadiabatic_dynamics, omega_pa = compute_postadiabatic_dynamics(
            omega_ref,
            omega_start,
            H,
            RR,
            chi1_v,
            chi2_v,
            m_1,
            m_2,
            splines,
            t_pn,
            dynamics_pn,
            tol=1e-12,
            params=params,
            order=8,
            postadiabatic_type=postadiabatic_type,
            window_length=10,
            only_first_n=12,
        )

        r0 = postadiabatic_dynamics[0, 1]
        pr0 = postadiabatic_dynamics[0, 3]
        pphi0 = postadiabatic_dynamics[0, 4]

        PA_success = True
    except Exception:
        # If PA fails use the adiabatic initial conditions
        r0, pphi0, pr0 = computeIC_augm(
            omega_start,
            H,
            RR,
            chi1_v,
            chi2_v,
            m_1,
            m_2,
            params=params,
        )
        params.p_params.omega = omega_start

    if PA_success:
        # Note that omega and omega_pa[0] may differ due to numerical noise
        # we set the initial frequency to omega_pa[0] to avoid interpolation errors
        params.p_params.omega = omega_pa[0]

    # Update parameters
    X1 = params.p_params.X_1
    X2 = params.p_params.X_2

    tmp = splines["everything"](params.p_params.omega)
    chi1_LN_start = tmp[0]
    chi2_LN_start = tmp[1]
    chi1_L_start = tmp[2]
    chi2_L_start = tmp[3]
    chi1_v_start = tmp[4:7]
    chi2_v_start = tmp[7:10]
    lN_start = tmp[10:13]

    params.p_params.chi1_v[:] = chi1_v_start
    params.p_params.chi2_v[:] = chi2_v_start

    params.p_params.chi_1, params.p_params.chi_2 = chi1_LN_start, chi2_LN_start
    params.p_params.chi1_L, params.p_params.chi2_L = chi1_L_start, chi2_L_start
    params.p_params.lN[:] = lN_start

    params.p_params.update_spins(chi1_LN_start, chi2_LN_start)

    ap_start = chi1_LN_start * X1 + chi2_LN_start * X2
    am_start = chi1_LN_start * X1 - chi2_LN_start * X2

    dSO_start = dSO_poly_fit(params.p_params.nu, ap_start, am_start)
    H.calibration_coeffs["dSO"] = dSO_start

    # Evaluate omega_circ and the Hamiltonian at the intial conditions
    q = np.array([r0, 0.0])
    p = np.array([pr0, pphi0])
    p_circ = np.array([0.0, p[1]])
    dynamics = H.dynamics(
        q,
        p,
        chi1_v_start,
        chi2_v_start,
        m_1,
        m_2,
        chi1_LN_start,
        chi2_LN_start,
        chi1_L_start,
        chi2_L_start,
    )
    H_val = dynamics[4]
    omega_circ = H.omega(
        q,
        p_circ,
        chi1_v_start,
        chi2_v_start,
        m_1,
        m_2,
        chi1_LN_start,
        chi2_LN_start,
        chi1_L_start,
        chi2_L_start,
    )
    params.p_params.omega_circ = omega_circ
    params.p_params.H_val = H_val

    # print(f"PA: r0 = {r0}, pphi0 = {pphi0}, pr0 = {pr0}")

    return r0, pphi0, pr0



from unittest.mock import patch

import numpy as np
import pandas as pd

from pyseobnr.eob.utils.utils import estimate_time_max_amplitude
from pyseobnr.eob.waveform.compute_hlms import compute_IMR_modes
from pyseobnr.generate_waveform import generate_modes_opt


def test_qnm_rotation_always_positive():
    """Checks we are applying the right rotation to the QNMs

    See https://git.ligo.org/waveforms/software/pyseobnr/-/merge_requests/70 for details
    """

    chi_1 = np.array((-0.0131941810847, 0.0676539715676, -0.646272606477))
    chi_2 = 0
    omega0 = 0.02

    with patch("pyseobnr.models.SEOBNRv5HM.compute_IMR_modes") as p_compute_IMR_modes:
        p_compute_IMR_modes.side_effect = compute_IMR_modes

        generate_modes_opt(
            q=5.33,
            chi1=chi_1,
            chi2=chi_2,
            omega_start=omega0,
            approximant="SEOBNRv5PHM",
        )

        p_compute_IMR_modes.assert_called_once()
        assert p_compute_IMR_modes.call_args.kwargs["qnm_rotation"] > 0

        p_compute_IMR_modes.reset_mock()

        generate_modes_opt(
            q=5.32,
            chi1=chi_1,
            chi2=chi_2,
            omega_start=omega0,
            approximant="SEOBNRv5PHM",
        )

        p_compute_IMR_modes.assert_called_once()
        assert p_compute_IMR_modes.call_args.kwargs["qnm_rotation"] > 0


def test_estimate_t_for_max_amplitude_PHM():
    """Checks that the estimated t_0 from the frame invariant amplitude is stable wrt. frequency"""
    chi_1 = np.array((-0.0131941810847, 0.0676539715676, -0.646272606477))
    chi_2 = 0.1
    omega0 = 0.02
    q = 10

    all_t0 = {}
    all_ts = {}
    all_frame_inv_amplitudes = {}

    with patch(
        "pyseobnr.models.SEOBNRv5HM.estimate_time_max_amplitude"
    ) as p_estimate_time_max_amplitude:

        def m_estimate_time_max_amplitude(*args, **kwargs):
            ret = estimate_time_max_amplitude(*args, **kwargs)
            # gets the outside freq
            all_t0[freq] = ret
            all_frame_inv_amplitudes[freq] = kwargs["amplitude"]
            return ret

        p_estimate_time_max_amplitude.side_effect = m_estimate_time_max_amplitude

        # 2**11, 2**12 ....
        freqs_to_check = np.logspace(10, 14, base=2, num=14 - 10 + 1)
        for freq in freqs_to_check:
            _, _, model = generate_modes_opt(
                q=q,
                chi1=chi_1,
                chi2=chi_2,
                omega_start=omega0,
                approximant="SEOBNRv5PHM",
                settings={"dt": 1 / float(freq), "M": 100},
                debug=True,
            )
            p_estimate_time_max_amplitude.assert_called_once()
            p_estimate_time_max_amplitude.reset_mock()

            all_ts[freq] = model.t

    # we calculate the value of the amplitude at 10M
    inter_at_10M = []
    for freq in freqs_to_check:
        inter_at_10M += [
            {
                "freq": freq,
                "t_0": all_t0[freq],
                # 10 -> this is 10M
                "val": np.interp(10, all_ts[freq], all_frame_inv_amplitudes[freq]),
            }
        ]

    inter_at_10M = pd.DataFrame(inter_at_10M)

    # the estimated t_0 does not "vary too much"
    assert inter_at_10M["t_0"].std() < 1e-2

    # the estimated amplitude at 10M does not vary too much
    # without the fix, we obtain 0.01928841154831644 for this specific setup
    assert inter_at_10M["val"].std() / inter_at_10M["val"].max() < 1e-3


def test_estimate_t_for_max_amplitude_HM():
    """Checks that the estimated t_0 from the frame invariant amplitude is stable wrt. frequency"""
    chi_1 = -0.646272606477
    chi_2 = 0.1
    omega0 = 0.02
    q = 10

    all_t0 = {}
    all_ts = {}
    all_frame_inv_amplitudes = {}

    with patch(
        "pyseobnr.models.SEOBNRv5HM.estimate_time_max_amplitude"
    ) as p_estimate_time_max_amplitude:

        def m_estimate_time_max_amplitude(*args, **kwargs):
            ret = estimate_time_max_amplitude(*args, **kwargs)
            # gets the outside freq
            all_t0[freq] = ret
            all_frame_inv_amplitudes[freq] = kwargs["amplitude"]
            return ret

        p_estimate_time_max_amplitude.side_effect = m_estimate_time_max_amplitude

        # 2**11, 2**12 ....
        freqs_to_check = np.logspace(10, 14, base=2, num=14 - 10 + 1)
        for freq in freqs_to_check:
            _, _, model = generate_modes_opt(
                q=q,
                chi1=chi_1,
                chi2=chi_2,
                omega_start=omega0,
                approximant="SEOBNRv5HM",
                settings={"dt": 1 / float(freq), "M": 100},
                debug=True,
            )
            p_estimate_time_max_amplitude.assert_called_once()
            p_estimate_time_max_amplitude.reset_mock()

            all_ts[freq] = model.t

    # we calculate the value of the amplitude at 10M
    inter_at_10M = []
    for freq in freqs_to_check:
        inter_at_10M += [
            {
                "freq": freq,
                "t_0": all_t0[freq],
                # 10 -> this is 10M
                "val": np.interp(10, all_ts[freq], all_frame_inv_amplitudes[freq]),
            }
        ]

    inter_at_10M = pd.DataFrame(inter_at_10M)

    # the estimated t_0 does in fact in this case vary so that stddev = 0.4108767856285839
    # assert inter_at_10M["t_0"].std() < 1e-2

    # the estimated amplitude at 10M does not vary too much
    # without the fix, we obtain 0.02275764373726919 for this specific setup
    assert inter_at_10M["val"].std() / inter_at_10M["val"].max() < 1e-3



from abc import ABC


class Model(ABC):
    """An abstract base class representing a model.
    Should serve as basis for concrete implementation
    """

    def __init__(self) -> None:
        pass

    def __call__(self):
        """Evaluate the waveform model"""
        pass



import numpy as np
import matplotlib.pyplot as plt
from matplotlib.lines import Line2D
import sys
import os
import importlib
import argparse

from scipy.interpolate import InterpolatedUnivariateSpline
from bilby.core.utils import check_directory_exists_and_if_not_mkdir
from pathos.multiprocessing import ProcessingPool as Pool
from pyseobnr.generate_waveform import generate_modes_opt

mode_list_v5HM = [(2, 2), (2, 1), (3, 3), (4, 4), (5, 5), (4, 3), (3, 2)]


def plots_smoothness_in_q(N: int, mode):
    """
    Plots of the (ell,m) mode amplitude and frequency for equal-spins varying q to show smoothness
    """

    colors = plt.cm.inferno(
        np.linspace(0, 1, N + 15)
    )  # adjust based on how many curves are plotted

    q_arr = np.linspace(1.0, 100.0, N)
    omega0 = 0.015

    ell, m = mode
    for chi in [
        -0.995,
        -0.5,
        0.0,
        0.5,
        0.995,
    ]:

        # amplitude plots
        for i, q in enumerate(q_arr):
            if not (m % 2 == 1 and q < 1.01):
                _, _, model = generate_modes_opt(q, chi, chi, omega0, debug=True)
                tm = model.t[np.argmax(np.abs(model.waveform_modes[f"2,2"]))]
                plt.plot(
                    model.t - tm,
                    np.abs(model.waveform_modes[f"{ell},{m}"]),
                    color=colors[i],
                )
        plt.xlim(-300, 100)
        plt.ylim(1e-5, 1e0)
        plt.yscale("log")
        plt.xlabel("$t/M$")
        plt.ylabel(f"$|h_{{{ell}{m}}}|$")
        plt.title(f"$\chi_1 = \chi_2 = {chi}$")
        legend_elements = [
            Line2D([0], [0], color=colors[0], lw=2, label=f"$q={q_arr[0]}$"),
            Line2D(
                [0], [0], color=colors[len(q_arr) - 1], lw=2, label=f"$q={q_arr[-1]}$"
            ),
        ]
        plt.legend(handles=legend_elements, loc="best")
        plt.savefig(
            f"./plots_smoothness/smoothness_{ell}{m}_amp_in_q_chi_{chi}.png", dpi=150
        )
        plt.close()

        # frequency plots
        for i, q in enumerate(q_arr):
            if not (m % 2 == 1 and q < 1.01):
                _, _, model = generate_modes_opt(q, chi, chi, omega0, debug=True)
                tm = model.t[np.argmax(np.abs(model.waveform_modes[f"2,2"]))]
                phi = -np.unwrap(np.angle(model.waveform_modes[f"{ell},{m}"]))
                iphi = InterpolatedUnivariateSpline(model.t, phi)
                omega_ellm = iphi.derivative()(model.t)
                plt.plot(model.t - tm, omega_ellm, color=colors[i])
        plt.xlim(-300, 100)
        id = np.abs(model.t - 100).argmin()
        lim = omega_ellm[id] + 0.2
        plt.ylim(0.0, lim)
        plt.xlabel("$t/M$")
        plt.ylabel(f"$\omega_{{{ell}{m}}}$")
        plt.title(f"$\chi_1 = \chi_2 = {chi}$")
        legend_elements = [
            Line2D([0], [0], color=colors[0], lw=2, label=f"$q={q_arr[0]}$"),
            Line2D(
                [0], [0], color=colors[len(q_arr) - 1], lw=2, label=f"$q={q_arr[-1]}$"
            ),
        ]
        plt.legend(handles=legend_elements, loc="best")
        plt.savefig(
            f"./plots_smoothness/smoothness_{ell}{m}_om_in_q_chi_{chi}.png", dpi=150
        )
        plt.close()


def plots_smoothness_in_chi(N: int, mode):
    """
    Plots of the (ell,m) mode amplitude and frequency for equal-spins varying chi to show smoothness
    """

    colors = plt.cm.inferno(
        np.linspace(0, 1, N + 15)
    )  # adjust based on how many curves are plotted

    chi_arr = np.linspace(-0.995, 0.995, N)
    omega0 = 0.015

    ell, m = mode
    #for q in [1.0, 2.0, 3.0, 4.0, 5.0, 8.0, 10.0, 15.0, 20.0, 30.0, 50.0, 75.0, 100.0]:
    for q in [1.0, 2.0, 8.0, 20.0]:

        # amplitude plots
        for i, chi in enumerate(chi_arr):
            if not (m % 2 == 1 and q < 1.01):
                _, _, model = generate_modes_opt(q, chi, chi, omega0, debug=True)
                tm = model.t[np.argmax(np.abs(model.waveform_modes[f"2,2"]))]
                plt.plot(
                    model.t - tm,
                    np.abs(model.waveform_modes[f"{ell},{m}"]),
                    color=colors[i],
                )
        plt.xlim(-300, 100)
        plt.ylim(1e-5, 1e0)
        plt.yscale("log")
        plt.xlabel("$t/M$")
        plt.ylabel(f"$|h_{{{ell}{m}}}|$")
        plt.title(f"$q = {q}$")
        legend_elements = [
            Line2D(
                [0], [0], color=colors[0], lw=2, label=f"$\chi_1 = \chi_2={chi_arr[0]}$"
            ),
            Line2D(
                [0],
                [0],
                color=colors[len(chi_arr) - 1],
                lw=2,
                label=f"$\chi_1 = \chi_2={chi_arr[-1]}$",
            ),
        ]
        plt.legend(handles=legend_elements, loc="best")
        plt.savefig(
            f"./plots_smoothness/smoothness_{ell}{m}_amp_in_chi_q_{q}.png", dpi=150
        )
        plt.close()

        # frequency plots
        for i, chi in enumerate(chi_arr):
            if not (m % 2 == 1 and q < 1.01):
                _, _, model = generate_modes_opt(q, chi, chi, omega0, debug=True)
                tm = model.t[np.argmax(np.abs(model.waveform_modes[f"2,2"]))]
                phi = -np.unwrap(np.angle(model.waveform_modes[f"{ell},{m}"]))
                iphi = InterpolatedUnivariateSpline(model.t, phi)
                omega_ellm = iphi.derivative()(model.t)
                plt.plot(model.t - tm, omega_ellm, color=colors[i])
                plt.xlim(-300, 100)
                id = np.abs(model.t - 100).argmin()
                lim = omega_ellm[id] + 0.2
                plt.ylim(0.0, lim)
        plt.xlabel("$t/M$")
        plt.ylabel(f"$\omega_{{{ell}{m}}}$")
        plt.title(f"$q = {q}$")
        legend_elements = [
            Line2D(
                [0], [0], color=colors[0], lw=2, label=f"$\chi_1 = \chi_2={chi_arr[0]}$"
            ),
            Line2D(
                [0],
                [0],
                color=colors[len(chi_arr) - 1],
                lw=2,
                label=f"$\chi_1 = \chi_2={chi_arr[-1]}$",
            ),
        ]

        plt.legend(handles=legend_elements, loc="best")
        plt.savefig(
            f"./plots_smoothness/smoothness_{ell}{m}_om_in_chi_q_{q}.png", dpi=150
        )
        plt.close()


def plot_one_mode(input):
    N, mode, quantity = input
    if quantity == "q":
        plots_smoothness_in_q(N, mode)
    elif quantity == "chi":
        plots_smoothness_in_chi(N, mode)


if __name__ == "__main__":

    p = argparse.ArgumentParser(
        description="Make plots to show smoothness varying physical parameters "
    )
    p.add_argument("--points", type=int, help="Number of points", default=41)
    p.add_argument("--n-cpu", type=int, help="Number of cpus to use", default=14)
    args = p.parse_args()

    check_directory_exists_and_if_not_mkdir("./plots_smoothness")

    lst_q = [(args.points, mode, "q") for mode in mode_list_v5HM]
    lst_chi = [(args.points, mode, "chi") for mode in mode_list_v5HM]
    lst = lst_q + lst_chi
    pool = Pool(args.n_cpu)
    pool.map(plot_one_mode, lst)




cimport cython
cimport numpy as np
import numpy as np

cdef class PhysicalParams:
    @cython.cdivision(True)
    @cython.embedsignature(True)
    def __cinit__(self,dict dc):
        self.m_1 = dc["m_1"]
        self.m_2 = dc["m_2"]
        self.M = self.m_1+self.m_2
        self.nu = self.m_1*self.m_2/(self.M)**2
        self.X_1 = self.m_1/self.M
        self.X_2 = self.m_2/self.M
        self.delta = self.m_1 - self.m_2
        self.a1 =  dc["a1"]
        self.a2 =  dc["a2"]
        self.chi_1 = dc["chi_1"]
        self.chi_2 = dc["chi_2"]
        self.chi1_v = dc["chi1_v"]
        self.chi2_v = dc["chi2_v"]
        self.lN = dc["lN"]
        self.omega = dc["omega"]
        self.omega_circ = dc["omega"]
        self.H_val = dc["H_val"]
        self._compute_derived_quants()

    cpdef update_spins(self,double chi_1,double chi_2):
        """
        Update the aligned spins and the derived quantities
        """
        self.chi_1 = chi_1
        self.chi_2 = chi_2
        self._compute_derived_quants()

    @cython.cdivision(True)
    cdef _compute_derived_quants(self):
        """
        Compute auxiliary spin quantities
        """
        # Symmetric and anti-symmetric spins
        self.chi_S = (self.chi_1+self.chi_2)/2
        self.chi_A = (self.chi_1-self.chi_2)/2
        # Test-particle spin
        self.a = (1 - 2 * self.nu) * self.chi_S + (self.m_1 - self.m_2) / (
            self.m_1 + self.m_2) * self.chi_A
        # a_{+,-}
        self.ap = self.X_1*self.chi_1+self.X_2*self.chi_2
        self.am = self.X_1*self.chi_1-self.X_2*self.chi_2

cdef class CalibCoeffs():
    def __cinit__(self,dc):
        self.dc = dc

    def __getitem__(self,str x):
        return self.dc[x]
    def __setitem__(self,str x,double y):
        self.dc[x] = y

cdef class FluxParams:
    @cython.embedsignature(True)
    def __cinit__(self,special_modes,extra_PN_terms):
        self.Tlm = np.zeros((ell_max+1,ell_max+1),dtype=np.float64)
        self.rho_coeffs = np.zeros((ell_max+1,ell_max+1,PN_limit),dtype=np.float64)
        self.rho_coeffs_log = np.zeros((ell_max+1,ell_max+1,PN_limit),dtype=np.float64)
        self.f_coeffs = np.zeros((ell_max+1,ell_max+1,PN_limit),dtype=np.float64)
        self.f_coeffs_vh = np.zeros((ell_max+1,ell_max+1,PN_limit),dtype=np.complex128)
        self.delta_coeffs = np.zeros((ell_max+1,ell_max+1,PN_limit),dtype=np.complex128)
        self.delta_coeffs_vh = np.zeros((ell_max+1,ell_max+1,PN_limit),dtype=np.complex128)
        self.rholm = np.zeros((ell_max+1,ell_max+1),dtype=np.complex128)
        self.deltalm = np.zeros((ell_max+1,ell_max+1),dtype=np.complex128)
        self.prefixes = np.zeros((ell_max+1,ell_max+1),dtype=np.complex128)
        self.prefixes_abs = np.zeros((ell_max+1,ell_max+1),dtype=np.float64)
        self.nqc_coeffs = np.zeros((ell_max+1,ell_max+1,3),dtype=np.float64)
        self.extra_coeffs = np.zeros((ell_max+1,ell_max+1,PN_limit),dtype=np.float64)
        self.extra_coeffs_log = np.zeros((ell_max+1,ell_max+1,PN_limit),dtype=np.float64)
        self.special_modes = special_modes
        self.extra_PN_terms = extra_PN_terms

cdef class Dynamics:
    def __cinit__(self):
        self.p_circ = np.zeros(2,dtype=np.float64)


@cython.embedsignature(True)
cdef class EOBParams:
    """
    A class to hold various EOB related quantities
    """
    def __init__(self,physical_params,coeffs,aligned=True,extra_PN_terms=True,
                mode_array=[(2,2),(2,1),(3,3),(3,2),(4,4),(4,3),(5,5)],
                special_modes=[(2,1),(4,3),(5,5)]):
        # Needed to get the signature to embed
        pass

    def __cinit__(self,physical_params,coeffs,aligned=True,extra_PN_terms=True,
                mode_array=[(2,2),(2,1),(3,3),(3,2),(4,4),(4,3),(5,5)],
                special_modes=[(2,1),(4,3),(5,5)]):
        # Physical params (e.g. spins)
        self.p_params = PhysicalParams(physical_params)
        # Calibration coefficients
        #self.c_coeffs = CalibCoeffs(coeffs)
        # Flux/waveform quantities
        self.flux_params = FluxParams(special_modes,extra_PN_terms)
        # Dynamics related quantities
        self.dynamics = Dynamics()
        # Mode array to use for generating modes
        self.mode_array = mode_array
        self.aligned = aligned



# cython: language_level=3

cimport cython
from pyseobnr.eob.utils.containers cimport EOBParams

cdef class Hamiltonian_C:
    cdef public EOBParams EOBpars
    cpdef _call(self, double[:]q,double[:]p,double chi_1,double chi_2,double m_1,double m_2)
    cpdef grad(self, double[:]q,double[:]p,double chi_1,double chi_2,double m_1,double m_2)
    cpdef hessian(self, double[:]q,double[:]p,double chi_1,double chi_2,double m_1,double m_2)
    cpdef double csi(self, double[:]q,double[:]p,double chi_1,double chi_2,double m_1,double m_2)
    cpdef dynamics(self, double[:]q,double[:]p,double chi_1,double chi_2,double m_1,double m_2)
    cpdef double omega(self, double[:]q,double[:]p,double chi_1,double chi_2,double m_1,double m_2)
    cpdef auxderivs(self, double[:]q,double[:]p,double chi_1,double chi_2,double m_1,double m_2)


import numpy as np
import pandas as pd


def compare_frames(
    test_frame, reference_frame, known_differences_percentage, time_tolerance_percent=1
):
    # sanity checks
    # all returned values are numbers
    assert test_frame.isna().any(axis=None).item() is False
    # all time steps are unique
    assert test_frame.t.is_unique and reference_frame.t.is_unique
    # time starts with 0
    assert test_frame.t[0] == 0
    # time is all positive
    assert (test_frame.t[1:] > 0).all()
    # time is ordered and strictly increasing
    assert (test_frame.t.diff()[1:] > 0).all()
    # time steps are decreasing non monotonically
    # this is not true
    # assert (frame_ehm.t.diff()[1:].diff().diff() < 0).all()

    # we generate the same number of elements in the dynamic
    assert len(reference_frame.columns) == reference_frame.shape[1]

    # dynamics stop at more or less the same time
    # this is a very rough check that values are not completely crazy
    t1_end = test_frame[["t"]].max().item()
    t2_end = reference_frame[["t"]].max().item()

    assert abs(t1_end - t2_end) / max(t1_end, t2_end) < (time_tolerance_percent / 100)

    # we create a new index composite of both index, keep unique values
    new_index_t = np.array(
        sorted(pd.concat((reference_frame.t, test_frame.t)).unique())
    )

    # we now inject the time steps of the reference frame into the current frame
    test_frame_missing_values = test_frame.set_index("t").reindex(new_index_t)
    reference_frame_missing_values = reference_frame.set_index("t").reindex(new_index_t)

    # we interpolate the missing values (method is important, index takes the time
    # point for performing the interpolation but is not precise enough)
    test_frame_missing_values_interpolated = test_frame_missing_values.interpolate(
        # method="index"
        method="cubic",
        order=3,
        limit_area="inside",
    )

    reference_frame_missing_values_interpolated = reference_frame_missing_values.interpolate(
        # method="index"
        method="cubic",
        order=3,
        limit_area="inside",
    )

    # we keep the section common to both
    test_frame_projected = test_frame_missing_values_interpolated[
        test_frame_missing_values_interpolated.index <= min(t1_end, t2_end)
    ]
    reference_frame_projected = reference_frame_missing_values_interpolated[
        reference_frame_missing_values_interpolated.index <= min(t1_end, t2_end)
    ]

    # all timesteps of both reference and test frames have now a proper value and they align
    assert test_frame_projected.isna().any(axis=None).item() is False
    assert reference_frame_projected.isna().any(axis=None).item() is False
    assert (test_frame_projected.index == reference_frame_projected.index).all()

    for idx_column, column in enumerate(reference_frame_projected.columns):
        percentage_tolerance = known_differences_percentage[column]

        diffs_column = (
            reference_frame_projected[[column]] - test_frame_projected[[column]]
        ).abs()

        fractional_diffs_column = (
            100
            * diffs_column[column]
            / pd.concat(
                (reference_frame_projected[column], test_frame_projected[column]),
                axis=1,
            ).max(axis=1)
        )

        assert fractional_diffs_column.max().item() < percentage_tolerance, (
            f"column {column} exceeds fractional tolerance: "
            f"%g > {percentage_tolerance} at index %d and time %g "
            "(reference=%g, test=%g)"
        ) % (
            fractional_diffs_column.max().item(),
            fractional_diffs_column.argmax(),
            fractional_diffs_column.idxmax().item(),
            reference_frame_projected.iloc[fractional_diffs_column.argmax()][
                [column]
            ].item(),
            test_frame_projected.iloc[fractional_diffs_column.argmax()][
                [column]
            ].item(),
        )



import argparse
import csv
import importlib
import os
import sys
from typing import List, Tuple

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
from bilby.core.utils import check_directory_exists_and_if_not_mkdir
from pathos.multiprocessing import ProcessingPool as Pool

from pyseobnr.auxiliary.sanity_checks.metrics  import UnfaithfulnessModeByModeLAL
from pyseobnr.auxiliary.external_models  import NRModel_SXS,SEOBNRv4HM_LAL
from pyseobnr.generate_waveform import generate_modes_opt

modes_v5HM = [(2, 2), (2, 1), (3, 3), (3, 2), (4, 4), (4, 3), (5, 5)]
modes_v4HM = [(2, 2), (2, 1), (3, 3), (4, 4), (5, 5)]

# modes_v5HM=[(2,2)]


def get_NR_paths(cases_file: str) -> List[str]:
    NR_path = []
    with open(cases_file, "r") as file:
        reader = csv.reader(file)
        next(reader, None)
        for row in reader:
            NR_path.append(row[5] + "/rhOverM_Asymptotic_GeometricUnits_CoM.h5")

    return NR_path


def single_mismatch(p: Tuple[str, str, bool]) -> Tuple:
    """Compute mode-by-mode mismatches between NR and the given EOB model

    Args:
        p (Tuple[str, str, bool]): The path to NR file, the type of model to use,
        whether to limit the mismatch computation to the merger-ringdown

    Returns:
        Tuple: Parameters and the mismatches, stored as a dict
    """
    path, model = p

    target_model = NRModel_SXS(path)
    q = target_model.q
    if q < 1:
        q = 1 / q
    if model == "SEOBNRv5HM":

        _, _, calib_model = generate_modes_opt(
            q, target_model.chi1[-1], target_model.chi2[-1], 0.7 * target_model.omega0,debug=True
        )
        modes = modes_v5HM
    elif model == "SEOBNRv4HM":
        calib_model = SEOBNRv4HM_LAL(
            q,
            target_model.chi1[-1],
            target_model.chi2[-1],
            0.7 * target_model.omega0,
        )
        calib_model()
        modes = modes_v4HM

    else:
        # print(f"Here, model={model}")
        calib_model = model

    masses = np.arange(10, 310, 10)
    unf_settings = {"debug": True, "masses": masses}
    unf = UnfaithfulnessModeByModeLAL(settings=unf_settings)

    mismatch = {}
    for mode in modes:
        mismatch[mode] = {}

    for mode in modes:
        ell, m = mode

        params = dict(
            q=target_model.q,
            chi1=np.array([0.0, 0.0, target_model.chi1[-1]]),
            chi2=np.array([0.0, 0.0, target_model.chi2[-1]]),
            omega0=0.7 * target_model.omega0,
            dt=target_model.delta_T,
            df=1 / (len(target_model.t) * target_model.delta_T),
        )
        #print(f"ell={ell},m={m}")
        mm = unf(target_model, calib_model, params=params, ell=ell, m=m)

        mismatch[mode] = mm

    q = np.round(target_model.q, 2)
    chi1 = np.round(target_model.chi1[-1], 3)
    chi2 = np.round(target_model.chi2[-1], 3)
    label = f"q_{q}_s{chi1}_s{chi2}"
    return label, q, chi1, chi2, mismatch


def mismatch_NR(
    NR_paths: List[str], n_cpu: int = 80, model: str = "SEOBNRv5HM"
) -> Tuple:
    """Compute mismatches between NR and a given model, in parallel for many
    cases.

    Args:
        NR_paths (List[str]): List of NR waveforms
        n_cpu (int, optional): Number of CPUs to use. Defaults to 80.
        model (str, optional): Model to compare to NR. Defaults to "SEOBNRv5HM".

    Returns:
        Tuple: Labels,parameters,mismatches. The mismatches are a dict, with
                with keys being the modes
    """

    mismatch = {}
    q_arr = []
    chi1_arr = []
    chi2_arr = []
    labels = []

    pool = Pool(n_cpu)
    lst = list(pool.map(single_mismatch, [(x, model) for x in NR_paths]))

    if model == "SEOBNRv5HM":
        modes = modes_v5HM
    elif model == "SEOBNRv4HM":
        modes = modes_v4HM
    else:
        modes = [(2, 2)]
    # Reshuffle the data into a convenient form
    # modes = [(2,2)]
    for mode in modes:
        mismatch[mode] = []
    for row in lst:
        label, q, chi1, chi2, mm = row
        q_arr.append(q)
        chi1_arr.append(chi1)
        chi2_arr.append(chi2)
        labels.append(label)
        for mode in modes:
            mismatch[mode].append(mm[mode])

    for mode in modes:
        mismatch[mode] = np.array(mismatch[mode])
    return (
        np.array(labels),
        np.array(q_arr),
        np.array(chi1_arr),
        np.array(chi2_arr),
        mismatch,
    )


if __name__ == "__main__":
    p = argparse.ArgumentParser()
    p.add_argument(
        "--cases-file", type=str, help="File containing the NR case specifications"
    )

    p.add_argument(
        "--mismatch-file", type=str, help="File with mismatches, if precomputed"
    )
    p.add_argument(
        "--model",
        type=str,
        help="Which approximant to compare to NR",
        default="SEOBNRv5HM",
    )
    p.add_argument(
        "--plots", action="store_true", help="Make diagnostic plots", default=True
    )
    p.add_argument(
        "--plots-only",
        action="store_true",
        help="Only generate plots, don't recompute things",
    )
    p.add_argument("--n-cpu", type=int, help="Number of cores to use", default=64)
    p.add_argument(
        "--include-all",
        action="store_true",
        help="Include odd m modes close to equal mass",
        default=False,
    )

    args = p.parse_args()

    if not args.plots_only:
        NR_paths = get_NR_paths(args.cases_file)

        # Get the mismatches
        labels, qs, chi1s, chi2s, mms = mismatch_NR(
            NR_paths,
            model=args.model,
            n_cpu=args.n_cpu,
        )

        # Save the parameters in a separate file, for convenience
        params = np.c_[qs, chi1s, chi2s]
        np.savetxt(f"parameters_mismatch_calc_{args.model}.dat", params)

        # Now save the mismatches, one file per mode
        for key in mms.keys():
            ell, m = key
            np.savetxt(f"mismatch_{args.model}{ell}{m}.dat", mms[key])

    if args.plots:
        params = np.genfromtxt(f"parameters_mismatch_calc_{args.model}.dat")
        q = params[:, 0]
        chi1 = params[:, 1]
        chi2 = params[:, 2]

        nu = q / (1 + q) ** 2
        m1 = q / (1 + q)
        m2 = 1 / (1 + q)
        ap = m1 * chi1 + m2 * chi2
        am = m1 * chi1 - m2 * chi2

        plt_dir = "./plots"
        check_directory_exists_and_if_not_mkdir(plt_dir)
        if args.model == "SEOBNRv5HM":
            mode_list = modes_v5HM
        elif args.model == "SEOBNRv4HM":
            mode_list = modes_v4HM
        else:

            mode_list = [(2, 2)]

        for mode in mode_list:
            ell, m = mode
            mm_M = np.loadtxt(f"mismatch_{args.model}{ell}{m}.dat")

            if not args.include_all:
                if m % 2:
                    idx_notEMES = np.where((q > 1.01) | (np.abs(chi1 - chi2) > 0.01))[0]
                    q_pl = q[idx_notEMES]
                    ap_pl = ap[idx_notEMES]
                    mm_M = mm_M[idx_notEMES]
                else:
                    q_pl = q
                    ap_pl = ap
            else:
                q_pl = q
                ap_pl = ap


            if True:
                M = np.arange(10, 310, 10)

                # maximum of mm_M across total mass for histogram
                mm = []
                for mp in mm_M:
                    mm.append(np.max(mp))
                mm = np.array(mm)

                # Spaghetti plot
                for mp in mm_M:
                    plt.plot(M, mp, color="C0", linewidth=0.5)
                plt.axhline(0.01, ls="--", color="red")
                plt.yscale("log")
                plt.xlabel("M")
                plt.ylabel("$\mathcal{M}$")
                plt.xlim(10, 300)
                plt.savefig(
                    f"{plt_dir}/mm_spaghetti_{args.model}{ell}{m}.png",
                    bbox_inches="tight",
                    dpi=300,
                )
                plt.close()

            # Histogram
            plt.hist(
                mm,
                bins=np.logspace(start=np.log10(0.00001), stop=np.log10(1.0), num=50),
                alpha=0.4,
                label=args.model,
            )
            plt.axvline(np.median(mm), c="C0", ls="--")
            plt.legend(loc="best")
            plt.gca().set_xscale("log")
            plt.xlabel("$\mathcal{M}_{\mathrm{Max}}$")
            plt.title(
                "$\mathcal{M}_{\mathrm{median}} = $" + f"{np.round(np.median(mm),6)}"
            )
            plt.savefig(
                f"{plt_dir}/mm_hist_{args.model}{ell}{m}.png",
                bbox_inches="tight",
                dpi=300,
            )
            plt.close()

            # CDF
            plt.hist(
                mm,
                bins=np.logspace(start=np.log10(0.00001), stop=np.log10(1.0), num=100),
                alpha=0.4,
                label=args.model,
                cumulative=True,
                density=True,
                histtype="step",
                lw=2,
            )

            plt.legend(loc="best")
            plt.gca().set_xscale("log")
            plt.xlabel("$\mathcal{M}_{\mathrm{Max}}$")
            plt.title(
                "$\mathcal{M}_{\mathrm{median}} = $" + f"{np.round(np.median(mm),6)}"
            )
            plt.grid(True, which="both", ls=":")
            plt.savefig(
                f"{plt_dir}/mm_cdf_{args.model}{ell}{m}.png",
                bbox_inches="tight",
                dpi=300,
            )
            plt.close()

            # Mismatch across parameter space
            mm_s, q_s, ap_s = map(list, zip(*sorted(zip(mm, q_pl, ap_pl))))
            plt.scatter(
                q_s, ap_s, c=mm_s, linewidths=1, norm=matplotlib.colors.LogNorm()
            )
            plt.ylabel("$\chi_{\mathrm{eff}}$")
            plt.xlabel("$q$")
            cbar = plt.colorbar()
            cbar.set_label("$\mathcal{M}$")
            plt.savefig(
                f"{plt_dir}/mm_scatter_{args.model}{ell}{m}.png",
                bbox_inches="tight",
                dpi=300,
            )
            plt.close()



from setuptools import setup, Extension
from Cython.Build import cythonize
from setuptools import find_packages
import numpy as np

# see https://cython.readthedocs.io/en/latest/src/userguide/migrating_to_cy30.html#numpy-c-api
# for the NPY_NO_DEPRECATED_API macro definition.

_numpy_no_deprecated_api = ("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")

extensions = [
    Extension(
        "pyseobnr.eob.utils.containers",
        ["pyseobnr/eob/utils/containers.pyx"],
        include_dirs=[np.get_include()],
        define_macros=[_numpy_no_deprecated_api],
    ),
    Extension(
        "pyseobnr.eob.waveform.waveform",
        ["pyseobnr/eob/waveform/waveform.pyx"],
        include_dirs=[np.get_include(), "pyseobnr/eob/utils"],
        define_macros=[_numpy_no_deprecated_api],
    ),
    Extension(
        "pyseobnr.eob.hamiltonian.Ham_align_a6_apm_AP15_DP23_gaugeL_Tay_C",
        ["pyseobnr/eob/hamiltonian/Ham_align_a6_apm_AP15_DP23_gaugeL_Tay_C.pyx"],
        include_dirs=[np.get_include(), "pyseobnr/eob/utils"],
        define_macros=[_numpy_no_deprecated_api],
    ),
    Extension(
        "pyseobnr.eob.hamiltonian.Hamiltonian_C",
        ["pyseobnr/eob/hamiltonian/Hamiltonian_C.pyx"],
        include_dirs=[np.get_include(), "pyseobnr/eob/utils"],
        define_macros=[_numpy_no_deprecated_api],
    ),
    Extension(
        "pyseobnr.eob.dynamics.rhs_aligned",
        ["pyseobnr/eob/dynamics/rhs_aligned.pyx"],
        include_dirs=[np.get_include(), "pyseobnr/eob/utils"],
        define_macros=[_numpy_no_deprecated_api],
    ),
    Extension(
        "pyseobnr.eob.dynamics.rhs_precessing",
        ["pyseobnr/eob/dynamics/rhs_precessing.pyx"],
        include_dirs=[np.get_include(), "pyseobnr/eob/utils"],
        define_macros=[_numpy_no_deprecated_api],
    ),
    Extension(
        "pyseobnr.eob.dynamics.postadiabatic_C",
        ["pyseobnr/eob/dynamics/postadiabatic_C.pyx"],
        include_dirs=[
            np.get_include(),
            "pyseobnr/eob/utils",
            "pyseobnr/eob/hamiltonian",
        ],
        define_macros=[_numpy_no_deprecated_api],
    ),
    Extension(
        "pyseobnr.eob.dynamics.postadiabatic_C_prec",
        ["pyseobnr/eob/dynamics/postadiabatic_C_prec.pyx"],
        include_dirs=[
            np.get_include(),
            "pyseobnr/eob/utils",
            "pyseobnr/eob/hamiltonian",
        ],
        define_macros=[_numpy_no_deprecated_api],
    ),
    Extension(
        "pyseobnr.eob.dynamics.postadiabatic_C_fast",
        ["pyseobnr/eob/dynamics/postadiabatic_C_fast.pyx"],
        include_dirs=[
            np.get_include(),
            "pyseobnr/eob/utils",
            "pyseobnr/eob/hamiltonian",
        ],
        define_macros=[_numpy_no_deprecated_api],
    ),
    Extension(
        "pyseobnr.eob.hamiltonian.Hamiltonian_v5PHM_C",
        ["pyseobnr/eob/hamiltonian/Hamiltonian_v5PHM_C.pyx"],
        include_dirs=[
            np.get_include(),
            "pyseobnr/eob/utils",
            "pyseobnr/eob/hamiltonian",
        ],
        define_macros=[_numpy_no_deprecated_api],
    ),
    Extension(
        "pyseobnr.eob.hamiltonian.Ham_AvgS2precess_simple_cython_PA_AD",
        ["pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx"],
        include_dirs=[
            np.get_include(),
            "pyseobnr/eob/utils",
            "pyseobnr/eob/hamiltonian",
        ],
        define_macros=[_numpy_no_deprecated_api],
    ),
]


setup(
    packages=find_packages(),
    ext_modules=cythonize(extensions),
    zip_safe=False,
)



import argparse

import matplotlib.pyplot as plt
import numpy as np
from bilby.core.utils import check_directory_exists_and_if_not_mkdir
from pathos.multiprocessing import ProcessingPool as Pool


from pyseobnr.generate_waveform import generate_modes_opt
from pyseobnr.auxiliary.sanity_checks.parameters import parameters_random_fast
from pyseobnr.auxiliary.sanity_checks.single_waveform_tests import amplitude_hierarchy_test


def amplitude_hierarchy(q: float, chi1: float, chi2: float, plots: bool = True):

    """
    Generate a SEOBNRv5 waveform with parameters (q, chi1, chi2) and run
    amplitude_hierarchy_test to check that the amplitude of the subdominant
    modes is less than the amplitude of the (2,2) mode until the peak of the (2,2) modes

    Return (0,0,0) if the test passed, (q, chi1, chi2) otherwise
    """

    omega0 = 0.015

    try:

        _, _, model = generate_modes_opt(q, chi1, chi2, omega0,debug=True)
        t_h = model.t
        hlms_dict = model.waveform_modes
        habs22 = np.abs(model.waveform_modes["2,2"])
        t_merger = t_h[np.argmax(habs22)]

        try:
            amplitude_hierarchy_test(t_h, hlms_dict)
            return (0, 0, 0)
        except AssertionError:
            print(
                f"Amplitude of one subdominant mode is higher than the (2,2) mode the following parameters: q = {q}, chi1 = {chi1}, chi2 = {chi2}"
            )
            if plots:
                label = f"q{np.round(q,2)}_s{np.round(chi1,3)}_s{np.round(chi2,3)}"
                plt_dir = "./plots_hierarchy"
                check_directory_exists_and_if_not_mkdir(plt_dir)

                for key in model.modes_list:
                    plt.plot(
                        t_h, np.abs(hlms_dict[f"{key[0]},{key[1]}"]), label=f"[{key}]"
                    )

                plt.xlabel("t/M")
                plt.legend(loc="best")
                plt.axvline(t_merger, ls="--")
                plt.xlim(t_merger - 30, t_merger + 30)
                plt.xlabel("$t/M$")
                plt.ylabel("$|h_{\ell m}|$")
                plt.title(f"{label}")
                plt.savefig(f"{plt_dir}/check_amplitude_hierarchy_{label}.png", dpi=150)
                plt.close()
            return (q, chi1, chi2)
    except IndexError:
        print(
            f"Error for the following parameters: q = {q}, chi1 = {chi1}, chi2 = {chi2}"
        )


def process_one_case(input):
    q, chi1, chi2, plots = input
    (
        q_bad,
        chi1_bad,
        chi2_bad,
    ) = amplitude_hierarchy(q, chi1, chi2, plots)
    return np.array([q_bad, chi1_bad, chi2_bad])


if __name__ == "__main__":

    p = argparse.ArgumentParser(
        description="Check amplitude hierarchy across parameter space"
    )
    p.add_argument("--points", type=int, help="Number of points", default="5000")
    p.add_argument("--q-max", type=float, help="Upper limit on q", default="100.0")
    p.add_argument(
        "--chi-max", type=float, help="Upper limit on |chi|", default="0.995"
    )
    p.add_argument(
        "--name",
        type=str,
        help="Name of the output file",
        default="amplitude_hierarchy.dat",
    )
    p.add_argument(
        "--plots", action="store_true", help="Make diagnostic plots", default=True
    )

    p.add_argument("--n-cpu", type=int, help="Number of cores to use", default=64)
    args = p.parse_args()

    qarr, chi1arr, chi2arr = parameters_random_fast(
        args.points,
        1.0,
        args.q_max,
        -args.chi_max,
        args.chi_max,
        -args.chi_max,
        args.chi_max,
    )

    lst = [(a, b, c, args.plots) for a, b, c in zip(qarr, chi1arr, chi2arr)]

    pool = Pool(args.n_cpu)

    all_means = pool.map(process_one_case, lst)

    all_means = np.array(all_means)
    np.savetxt(args.name, all_means)

    ### Plots ###

    if args.plots:

        plt_dir = "./plots_hierarchy"
        check_directory_exists_and_if_not_mkdir(plt_dir)

        res_path = args.name
        res = np.loadtxt(res_path)

        q = res[:, 0]
        idx = np.where(q > 0.0)[0]

        q = q[idx]
        chi1 = res[:, 1][idx]
        chi2 = res[:, 2][idx]

        nu = q / (1 + q) ** 2
        m1 = q / (1 + q)
        m2 = 1 / (1 + q)
        ap = m1 * chi1 + m2 * chi2
        am = m1 * chi1 - m2 * chi2

        if len(q) > 0:
            # Cases where modes amplitude hierarchy is not satisfied across parameter space - (q, chieff)
            plt.scatter(q, ap)
            plt.ylabel("$\chi_{\mathrm{eff}}$")
            plt.xlabel("$q$")

            plt.title("Cases where modes amplitude_hierarchy is not satisfied")
            plt.savefig(
                f"{plt_dir}/amplitude_hierarchy.png", bbox_inches="tight", dpi=300
            )
            plt.close()
        else:
            print("Test passed for all cases")



"""
Contains hand-written vector operations wrapped in numba for speed.
For small operations, i.e. involving just 2 vectors these are faster
than NumPy since they do not incur overheads.
"""

from math import sqrt

import numpy as np
from numba import float64, jit


@jit(float64[:](float64[:], float64[:]), cache=True, nopython=True)
def my_cross(a, b):
    """
    Function to compute the cross product between two 3D arrays

    Args:
        a (np.ndarray): Array a
        b (np.ndarray): Array b

    Returns:
        (np.ndarray): Cross product axb
    """
    result = np.empty(3, dtype=np.float64)
    result[0] = a[1] * b[2] - a[2] * b[1]
    result[1] = a[2] * b[0] - a[0] * b[2]
    result[2] = a[0] * b[1] - a[1] * b[0]
    return result


@jit(float64(float64[:], float64[:]), cache=True, nopython=True)
def my_dot(a, b):
    """
    Function to compute the dot product between two 3D arrays

    Args:
        a (np.ndarray): Array a
        b (np.ndarray): Array b

    Returns:
        (float): Dot product a.b
    """

    return a[0] * b[0] + a[1] * b[1] + a[2] * b[2]


@jit(
    float64(float64[:]),
    cache=True,
    nopython=True,
)
def my_norm(a):
    """
    Function to compute the norm of a 3D array

    Args:
        a (np.ndarray): Array a

    Returns:
        (float): L2-norm of a
    """
    norm = sqrt(a[0] * a[0] + a[1] * a[1] + a[2] * a[2])

    return norm
